#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements the basic resources. See builder_datastore for specific functions to initiate datastores.
"""
import threading
from threading import Semaphore
from typing import Any, Generator, Union, Set, List, Dict, Tuple
import io
import os
import glob
import fnmatch

import pandas as pd
import requests

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element
from ckanapi_harvesters.auxiliary.ckan_progress_callbacks import CkanProgressCallback, CkanCallbackLevel
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo
from ckanapi_harvesters.auxiliary.path import resolve_rel_path, glob_rm_glob, glob_name
from ckanapi_harvesters.builder.builder_errors import ResourceFileNotExistMessage
from ckanapi_harvesters.builder.builder_resource_multi_abc import BuilderMultiABC, FileChunkDataFrame
from ckanapi_harvesters.builder.builder_resource import BuilderResourceABC, initial_resource_building_state

multi_file_exclude_other_files:bool = True


class BuilderMultiFile(BuilderResourceABC, BuilderMultiABC):
    """
    Class to manage a set of files to upload as separate resources
    """
    def __init__(self, *, parent, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, dir_name:str=None):
        super().__init__(parent=parent, name=name, format=format, description=description, resource_id=resource_id, download_url=download_url)
        self.dir_name: str = dir_name
        self.local_file_list_base_dir: str = ""
        self.local_file_list: Union[List[str], None] = None
        self.local_file_size: Union[List[int], None] = None
        self.local_file_size_sum: Union[int, None] = None
        self.excluded_files: Set[str] = set()
        self.remote_resource_names: Union[List[str], None] = None
        self.excluded_resource_names: Set[str] = set()
        # BuilderMultiABC:
        self.stop_event = threading.Event()
        self.thread_ckan: Dict[str, CkanApi] = {}
        self.progress_callback = CkanProgressCallback()
        self.enable_multi_threaded_upload:bool = True
        self.enable_multi_threaded_download:bool = True
        self.file_semaphore = Semaphore()

    @staticmethod
    def resource_mode_str() -> str:
        return "MultiFile"

    def copy(self, *, dest=None, parent=None):
        if dest is None:
            dest = BuilderMultiFile(parent=self.parent_package)
        super().copy(dest=dest, parent=parent)
        dest.dir_name = self.dir_name
        # BuilderMultiABC:
        dest.progress_callback = self.progress_callback.copy()
        dest.enable_multi_threaded_upload = self.enable_multi_threaded_upload
        dest.enable_multi_threaded_download = self.enable_multi_threaded_download
        # do not copy stop_event
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row, base_dir=base_dir)
        self.dir_name = _string_from_element(row["file/url"], empty_value="", strip=True)

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = self.dir_name
        return d

    def get_or_query_resource_id(self, ckan: CkanApi, cancel_if_present:bool=True, error_not_found:bool=True) -> Union[None,str]:
        return None


    ## upload --------------------------------------------------------------------
    def patch_request(self, ckan: CkanApi, package_id: str, *, reupload: bool = None, override_ckan:bool=False,
                      resources_base_dir:str=None,
                      payload:Union[bytes, io.BufferedIOBase]=None) -> Union[None, CkanResourceInfo]:
        return None

    def upload_request_final(self, ckan:CkanApi, *, force:bool=False) -> None:
        return None

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return False

    def get_sample_file_path(self, resources_base_dir:str, ckan:Union[CkanApi,None]=None, file_index:int=0) -> Union[str,None]:
        self.list_local_files(resources_base_dir=resources_base_dir)
        return self.local_file_list[file_index]

    def load_sample_data(self, resources_base_dir:str, file_index:int=0) -> Union[bytes,None]:
        file_path:str = self.get_sample_file_path(resources_base_dir, file_index=file_index)
        with open(file_path, "rb") as f:
            return f.read()

    def list_local_files(self, resources_base_dir:str, cancel_if_present:bool=True,
                         excluded_files:Set[str]=None) -> Union[List[str],None]:
        """
        List files corresponding to the multi-file resource configuration and are not used in mono-resources

        :param resources_base_dir:
        :param cancel_if_present:
        :param excluded_files: files from mono-resources
        :return:
        """
        if excluded_files is None:
            excluded_files = set()
        if (cancel_if_present and self.local_file_list is not None
                and self.local_file_list_base_dir == resources_base_dir
                and self.excluded_files == excluded_files):
            return self.local_file_list
        dir_search_path = resolve_rel_path(resources_base_dir, self.dir_name, field=f"File/URL of resource {self.name}")
        search_query = dir_search_path
        file_set = set(glob.glob(search_query))
        file_set = file_set - excluded_files
        file_list = list(file_set)
        file_list.sort()
        self.local_file_list = file_list
        self.local_file_size = [0] * len(file_list)
        for index, file_name in enumerate(file_list):
            self.local_file_size[index] = os.path.getsize(file_name)
        self.local_file_size_sum = sum(self.local_file_size, 0)
        self.local_file_list_base_dir = resources_base_dir
        self.excluded_files = excluded_files
        return file_list

    def init_local_files_list(self, resources_base_dir:str, cancel_if_present:bool=True, excluded_files:Set[str]=None, **kwargs) -> List[str]:
        return self.list_local_files(resources_base_dir=resources_base_dir, cancel_if_present=cancel_if_present,
                                     excluded_files=excluded_files)

    def get_local_file_total_size(self) -> int:
        return self.local_file_size_sum

    def get_local_file_offset(self, file_chunk: FileChunkDataFrame) -> int:
        return sum(self.local_file_size[:file_chunk.file_index]) + file_chunk.file_position

    def get_local_file_count(self) -> int:
        if self.local_file_list is None:
            raise RuntimeError("You must call list_local_files first")
        return len(self.local_file_list)

    def get_local_file_generator(self, resources_base_dir:str, excluded_files:Set[str]=None, **kwargs) -> Generator[str, None, None]:
        self.list_local_files(resources_base_dir=resources_base_dir, excluded_files=excluded_files)
        for file_name in self.local_file_list:
            yield file_name

    def get_local_df_chunk_generator(self, resources_base_dir:str, ckan:CkanApi, excluded_files:Set[str]=None, **kwargs) -> Generator[FileChunkDataFrame, None, None]:
        for file_index, file_name in enumerate(self.get_local_file_generator(resources_base_dir=resources_base_dir)):
            yield FileChunkDataFrame(None, file_name, file_index, 0, 0, 0)

    def upload_file_checks(self, *, resources_base_dir: str = None, ckan: CkanApi = None, excluded_files:Set[str]=None, **kwargs) \
            -> Union[None, ContextErrorLevelMessage]:
        if os.path.isdir(resolve_rel_path(resources_base_dir, glob_rm_glob(self.dir_name), field=f"File/URL of resource {self.name}")):
            if len(self.list_local_files(resources_base_dir=resources_base_dir, excluded_files=excluded_files)) > 0:
                return None
            else:
                return ResourceFileNotExistMessage(self.name, ErrorLevel.Error,
                    f"Empty resource directory for multi-file resource {self.name}: {os.path.join(resources_base_dir, self.dir_name)}")
        else:
            return ResourceFileNotExistMessage(self.name, ErrorLevel.Error,
                f"Missing directory for multi-file resource {self.name}: {os.path.join(resources_base_dir, self.dir_name)}")

    def upload_file_chunk(self, ckan:CkanApi, package_id:str, file_chunk:FileChunkDataFrame, *,
                          reupload:bool=False, override_ckan:bool=False, cancel_if_present:bool=True) -> CkanResourceInfo:
        """
        Upload a file, using its name as resource name
        """
        # The FileStore implementation uploads each file as a new resource. Files are not read by chunks but are fully loaded in memory in binary mode.
        file_path = file_chunk.file_path
        _, resource_name = os.path.split(file_path)
        resource_info = ckan.map.get_resource_info(resource_name, package_name=package_id, error_not_mapped=False)
        self.known_resource_info = resource_info
        self.known_id = resource_info.id
        self._merge_resource_attributes(override_ckan=override_ckan)
        if resource_info is not None and cancel_if_present and not reupload:
            resource_info.newly_created = False
            resource_info.newly_updated = False
            return resource_info
        resource_info = ckan.resource_create(package_id, resource_name, format=self.resource_attributes.format, description=self.resource_attributes.description,
                                    state=initial_resource_building_state if initial_resource_building_state is not None else self.resource_attributes.state,
                                    file_path=file_path, reupload=reupload, cancel_if_exists=True, update_if_exists=True,
                                    create_default_view=True, auto_submit=False, progress_callback=self.progress_callback)
        self.upload_request_final(ckan)
        self.known_resource_info = None
        self.known_id = None
        return resource_info


    def _unit_upload_apply(self, *, ckan:CkanApi, file_chunk:FileChunkDataFrame,
                           overall_chunk_index:int, start_index:int, end_index:int, file_count:int,
                           package_id:str, reupload:bool, only_missing:bool, excluded_files:Set[str]) -> None:
        # For each file, this function initiates its own FileStore.
        file_path = file_chunk.file_path
        _, file_name = os.path.split(file_path)
        index = file_chunk.file_index
        if start_index <= index and index < end_index and file_path not in excluded_files:
            self.progress_callback.task_progress(self.get_local_file_offset(file_chunk), self.get_local_file_total_size(), info=file_path,
                                         file_index=file_chunk.file_index, file_count=self.get_local_file_count(),
                                         context=f"{ckan.identifier} upload", level=CkanCallbackLevel.MultiFileResource)
            self.upload_file_chunk(ckan=ckan, package_id=package_id, file_chunk=file_chunk,
                                   reupload=reupload, cancel_if_present=only_missing)
        else:
            # self.progress_callback.call(index, total, info=df_upload_local, context=f"{ckan.identifier} single-thread skip", level=CkanCallbackLevel.MultiFileResource)
            pass

    def upload_request_full(self, ckan:CkanApi, resources_base_dir:str, *,
                            threads:int=1, external_stop_event=None,
                            start_index:int=0, end_index:int=None, allow_chunks:bool=True,
                            reupload:bool=False, only_missing:bool=False, from_line_count:bool=False,
                            excluded_files:Set[str]=None) -> None:
        if excluded_files is None:
            excluded_files = set()
        package_id = self.get_or_query_package_id(ckan)
        super().upload_request_full(ckan=ckan, resources_base_dir=resources_base_dir, threads=threads,
                                    external_stop_event=external_stop_event, start_index=start_index, end_index=end_index,
                                    reupload=reupload, only_missing=only_missing, from_line_count=from_line_count,
                                    package_id=package_id, excluded_files=excluded_files, allow_chunks=allow_chunks)
        # if threads < 0:
        #     # cancel large uploads in this case
        #     return None
        # elif threads is None or threads > 1:
        #     return self.upload_request_full_multi_threaded(ckan=ckan, resources_base_dir=resources_base_dir,
        #                                                    threads=threads, external_stop_event=external_stop_event,
        #                                                    start_index=start_index, end_index=end_index,
        #                                                    reupload=reupload, only_missing=only_missing,
        #                                                    excluded_files=excluded_files)
        # else:
        #     self.init_local_files_list(resources_base_dir=resources_base_dir, cancel_if_present=True, excluded_files=excluded_files)
        #     package_id = self.get_or_query_package_id(ckan)
        #     if ckan.verbose_extra:
        #         print(f"Launching single-threaded upload of multi-file resource {self.name}")
        #     total = self.get_local_file_len()
        #     end_index = positive_end_index(end_index, total)
        #     for index, file_path in enumerate(self.get_local_file_generator(resources_base_dir=resources_base_dir, excluded_files=excluded_files)):
        #         if external_stop_event is not None and external_stop_event.is_set():
        #             print(f"{ckan.identifier} Interrupted")
        #             return
        #         self._unit_upload_apply(ckan, file=file_path, package_id=package_id,
        #                                 reupload=reupload, only_missing=only_missing,
        #                                 index=index, start_index=start_index, end_index=end_index, total=total,
        #                                 excluded_files=excluded_files)
        #     self._call_progress_callback(total, total, context=f"{ckan.identifier} single-thread upload")
        #     # at last, apply final actions:
        #     self.upload_request_final(ckan)

    # def upload_request_graceful(self, ckan:CkanApi, file_path: str, *, index:int, package_id:str,
    #                             external_stop_event=None,
    #                             start_index:int=0, end_index:int=None,
    #                             reupload:bool=False, only_missing:bool=False, excluded_files:Set[str]=None) -> None:
    #     """
    #     Calls upload_file with checks specific to multi-threading.
    #
    #     :return:
    #     """
    #     # ckan.session_reset()
    #     # ckan.identifier = current_thread().name
    #     ckan = self.thread_ckan[current_thread().name]
    #     total = self.get_local_file_len()
    #     end_index = positive_end_index(end_index, total)
    #     if self.stop_event.is_set():
    #         return
    #     if external_stop_event is not None and external_stop_event.is_set():
    #         print(f"{ckan.identifier} Interrupted")
    #         return
    #     try:
    #         self._unit_upload_apply(ckan, file=file_path, package_id=package_id,
    #                                 reupload=reupload, only_missing=only_missing,
    #                                 index=index, start_index=start_index, end_index=end_index, total=total,
    #                                 excluded_files=excluded_files)
    #     except Exception as e:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print(f"Stopping all threads because an exception occurred in thread: {e}")
    #         raise e from e

    # def upload_request_full_multi_threaded(self, ckan:CkanApi, resources_base_dir:str,
    #                         threads:int=1, external_stop_event=None,
    #                         start_index:int=0, end_index:int=None,
    #                         reupload:bool=False, only_missing:bool=False, excluded_files:Set[str]=None):
    #     """
    #     Multi-threaded implementation of upload_request_full, using ThreadPoolExecutor.
    #     """
    #     self.init_local_files_list(resources_base_dir=resources_base_dir, cancel_if_present=True, excluded_files=excluded_files)
    #     package_id = self.get_or_query_package_id(ckan)
    #     self._prepare_for_multithreading(ckan)
    #     try:
    #         with ThreadPoolExecutor(max_workers=threads, initializer=self._init_thread, initargs=(ckan,)) as executor:
    #             if ckan.verbose_extra:
    #                 print(f"Launching multi-threaded upload of multi-file resource {self.name}")
    #             futures = [executor.submit(self.upload_request_graceful, ckan=ckan, file_path=file_path, index=index, package_id=package_id,
    #                                        start_index=start_index, end_index=end_index, external_stop_event=external_stop_event,
    #                                        excluded_files=excluded_files, reupload=reupload, only_missing=only_missing)
    #                        for index, file_path in enumerate(self.get_local_file_generator(resources_base_dir=resources_base_dir, excluded_files=excluded_files))]
    #             for future in futures:
    #                 future.result()  # This will propagate the exception
    #         total = self.get_local_file_len()
    #         self._call_progress_callback(total, total, context=f"{ckan.identifier} multi-thread upload")
    #     except Exception as e:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print(f"Stopping all threads because an exception occurred: {e}")
    #         raise e from e
    #     finally:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print("End of multi-threaded upload...")
    #     # at last, apply final actions:
    #     self.upload_request_final(ckan)


    ## download ------------------------------------------------
    def list_remote_resources(self, ckan:CkanApi, *, excluded_resource_names:Set[str]=None,
                              cancel_if_present: bool = True) -> List[str]:
        """
        Defines the list of resources to download that correspond to the definition and are not used in mono-resources.

        :param ckan:
        :param excluded_resource_names: resource names of mono-resources
        :param cancel_if_present:
        :return:
        """
        if cancel_if_present and self.remote_resource_names is not None and self.excluded_resource_names == excluded_resource_names:
            return self.remote_resource_names
        if excluded_resource_names is None:
            excluded_resource_names = set()
        package_info = ckan.get_package_info_or_request(self.package_name)
        resource_names = set(package_info.resources_id_index.keys())
        # resource_name_glob = self.name
        resource_name_glob = glob_name(self.dir_name)
        filtered_resource_names = set(fnmatch.filter(resource_names, resource_name_glob))  # apply name as wildcard filter
        filtered_resource_names = filtered_resource_names - excluded_resource_names
        self.remote_resource_names = sorted(list(filtered_resource_names))
        self.excluded_resource_names = excluded_resource_names
        return self.remote_resource_names

    def list_remote_resource_ids(self, ckan:CkanApi, *, excluded_resource_names:Set[str]=None,
                              cancel_if_present: bool = True) -> List[str]:
        resource_names = self.list_remote_resources(ckan, excluded_resource_names=excluded_resource_names,
                                                    cancel_if_present=cancel_if_present)
        resource_ids = [ckan.map.get_resource_id(resource_name, package_name=self.package_name) for resource_name in resource_names]
        return resource_ids

    def init_download_file_query_list(self, ckan: CkanApi, out_dir: str=None,
                                      cancel_if_present: bool = True,
                                      excluded_resource_names:Set[str]=None, **kwargs) -> List[str]:
        if out_dir is not None:
            dir_tables = resolve_rel_path(out_dir, glob_rm_glob(self.dir_name, default_rec_dir=self.name), field=f"File/URL of resource {self.name}")
            os.makedirs(dir_tables, exist_ok=True)
        return self.list_remote_resources(ckan, cancel_if_present=cancel_if_present, excluded_resource_names=excluded_resource_names)

    def get_file_query_generator(self) -> Generator[str, Any, None]:
        for resource_name in self.remote_resource_names:
            yield resource_name

    def get_file_query_count(self) -> int:
        if self.remote_resource_names is None:
            raise RuntimeError("init_download_file_query_list must be called first")
        return len(self.remote_resource_names)

    def download_file_query_item(self, ckan: CkanApi, out_dir: str, file_query_item: str) \
            -> Tuple[Union[str,None], Union[requests.Response,None]]:
        resource_name = file_query_item
        file_out = None
        if out_dir is not None:
            file_out = resolve_rel_path(out_dir, glob_rm_glob(self.dir_name, default_rec_dir=self.name), resource_name, field=f"File/URL of resource {self.name}")
            if self.download_skip_existing and os.path.exists(file_out):
                if ckan.params.verbose_extra:
                    print(f"Skipping existing file {file_out}")
                return file_out, None
        resource_id = ckan.map.get_resource_id(resource_name, package_name=self.package_name)
        resource_info, response = ckan.resource_download(resource_id)
        if out_dir is not None:
            with open(file_out, 'wb') as f:
                f.write(response.content)
        else:
            file_out = None
        return file_out, response

    def download_request_generator(self, ckan: CkanApi, out_dir: str,
                                   excluded_resource_names:Set[str]=None) -> Generator[Tuple[Union[str,None], Union[requests.Response,None]], Any, None]:
        self.init_download_file_query_list(ckan=ckan, out_dir=out_dir, cancel_if_present=True,
                                           excluded_resource_names=excluded_resource_names)
        for file_query_item in self.get_file_query_generator():
            yield self.download_file_query_item(ckan=ckan, out_dir=out_dir, file_query_item=file_query_item)

    def _unit_download_apply(self, ckan:CkanApi, file_query_item:Any, out_dir:str,
                             index:int, start_index:int, end_index:int, total:int,
                             excluded_resource_names:Set[str]) -> Any:
        if start_index <= index and index < end_index and file_query_item not in excluded_resource_names:
            self.progress_callback.task_progress(index, total, info=file_query_item,
                                         context=f"{ckan.identifier} single-thread download", level=CkanCallbackLevel.MultiFileResource)
            self.download_file_query_item(ckan=ckan, out_dir=out_dir, file_query_item=file_query_item)
        else:
            pass
            # self.progress_callback.call(index, total, info=file_query_item, context=f"{ckan.identifier} single-thread skip", level=CkanCallbackLevel.MultiFileResource)

    def download_request_full(self, ckan: CkanApi, out_dir: str, threads:int=1, external_stop_event=None,
                              start_index:int=0, end_index:int=None, force:bool=False,
                              excluded_resource_names:Set[str]=None) -> None:
        return super().download_request_full(ckan=ckan, out_dir=out_dir, threads=threads,
                                             external_stop_event=external_stop_event,
                                             start_index=start_index, end_index=end_index, force=force,
                                             excluded_resource_names=excluded_resource_names)
        # if (not self.enable_download) and (not force):
        #     msg = f"Did not download resource {self.name} because download was disabled."
        #     warn(msg)
        #     return None
        # if threads < 0:
        #     # do not download large datasets in this case
        #     return None
        # elif threads is None or threads > 1:
        #     return self.download_request_full_multi_threaded(ckan=ckan, out_dir=out_dir,
        #                                                      threads=threads, external_stop_event=external_stop_event,
        #                                                      start_index=start_index, end_index=end_index,
        #                                                      excluded_resource_names=excluded_resource_names)
        # else:
        #     self.init_download_file_query_list(ckan=ckan, out_dir=out_dir, cancel_if_present=True,
        #                                        excluded_resource_names=excluded_resource_names)
        #     if ckan.verbose_extra:
        #         print(f"Launching single-threaded download of multi-file resource {self.name}")
        #     total = self.get_file_query_len()
        #     end_index = positive_end_index(end_index, total)
        #     for index, file_query_item in enumerate(self.get_file_query_generator()):
        #         if external_stop_event is not None and external_stop_event.is_set():
        #             print(f"{ckan.identifier} Interrupted")
        #             return
        #         self._unit_download_apply(ckan=ckan, file_query_item=file_query_item, out_dir=out_dir,
        #                                   index=index, start_index=start_index, end_index=end_index, total=total,
        #                                   excluded_resource_names=excluded_resource_names)
        #     self._call_progress_callback(total, total, context=f"{ckan.identifier} single-thread download")

    # def download_file_query_item_graceful(self, ckan: CkanApi, out_dir: str, resource_name: str, index:int,
    #                                       external_stop_event=None, start_index:int=0, end_index:int=None,
    #                                       excluded_resource_names:Set[str]=None) -> None:
    #     """
    #     Implementation of download_file_query_item with checks for a multi-threaded download.
    #     """
    #     # ckan.session_reset()
    #     # ckan.identifier = current_thread().name
    #     ckan = self.thread_ckan[current_thread().name]
    #     total = self.get_file_query_len()
    #     end_index = positive_end_index(end_index, total)
    #     if self.stop_event.is_set():
    #         return
    #     if external_stop_event is not None and external_stop_event.is_set():
    #         print(f"{ckan.identifier} Interrupted")
    #         return
    #     try:
    #         self._unit_download_apply(ckan=ckan, file_query_item=file_query_item, out_dir=out_dir,
    #                                   index=index, start_index=start_index, end_index=end_index, total=total,
    #                                   excluded_resource_names=excluded_resource_names)
    #     except Exception as e:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print(f"Stopping all threads because an exception occurred in thread: {e}")
    #         raise e from e

    # def download_request_full_multi_threaded(self, ckan: CkanApi, out_dir: str,
    #                                          threads: int = None, external_stop_event=None,
    #                                          start_index:int=0, end_index:int=-1,
    #                                          excluded_resource_names:Set[str]=None) -> None:
    #     """
    #     Multi-threaded implementation of download_request_full using ThreadPoolExecutor.
    #     """
    #     self.init_download_file_query_list(ckan=ckan, out_dir=out_dir, cancel_if_present=True, excluded_resource_names=excluded_resource_names)
    #     self._prepare_for_multithreading(ckan)
    #     try:
    #         with ThreadPoolExecutor(max_workers=threads, initializer=self._init_thread, initargs=(ckan,)) as executor:
    #             if ckan.verbose_extra:
    #                 print(f"Launching multi-threaded download of multi-file resource {self.name}")
    #             futures = [executor.submit(self.download_file_query_item_graceful, ckan=ckan, out_dir=out_dir, resource_name=resource_name,
    #                                        index=index, external_stop_event=external_stop_event, start_index=start_index, end_index=end_index,
    #                                        excluded_resource_names=excluded_resource_names)
    #                        for index, resource_name in enumerate(self.get_file_query_generator())]
    #             for future in futures:
    #                 future.result()  # This will propagate the exception
    #         total = self.get_file_query_len()
    #         self._call_progress_callback(total, total, context=f"multi-thread download")
    #     except Exception as e:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print(f"Stopping all threads because an exception occurred: {e}")
    #         raise e from e
    #     finally:
    #         self.stop_event.set()  # Ensure all threads stop
    #         if ckan.verbose_extra:
    #             print("End of multi-threaded download...")

    def download_resource_bytes(self, ckan:CkanApi, full_download:bool=True, **kwargs) -> Union[bytes, None]:
        return None

    def download_request(self, ckan: CkanApi, out_dir: str, *, full_download:bool=True, threads:int=1,
                         force:bool=False, excluded_resource_names:Set[str]=None, return_data:bool=False, **kwargs) -> None:
        if full_download:
            return self.download_request_full(ckan=ckan, out_dir=out_dir, threads=threads, force=force,
                                              excluded_resource_names=excluded_resource_names, **kwargs)

    def resource_info_request(self, ckan:CkanApi, error_not_found:bool=True) -> Union[CkanResourceInfo, None]:
        return None  # there are multiple resource ids => do not return info
    def _to_ckan_resource_info(self, package_id:str, check_id:bool=True) -> CkanResourceInfo:
        return None



#!python3
# -*- coding: utf-8 -*-
"""
Tests to perform after the example package was uploaded
"""
from ckanapi_harvesters.builder.builder_package import BuilderPackage
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_action import CkanNotFoundError, CkanSqlCapabilityError

from ckanapi_harvesters.builder.example import example_package_xls


def run(ckan:CkanApi = None):
    BuilderPackage.unlock_external_code_execution()  # comment to test if the safety feature is enabled

    mdl = BuilderPackage.from_excel(example_package_xls)
    ckan = mdl.init_ckan(ckan)
    ckan.initialize_from_cli_args()
    ckan.input_missing_info(input_args_if_necessary=True, input_owner_org=True)
    ckan.set_verbosity(True)

    # ckan.apikey.clear()  # uncomment for specific test

    capability = ckan.test_sql_capabilities()
    try:
        ckan.api_help_show("datastore_search_sql")
    except CkanNotFoundError:
        print("No datastore_search_sql help")

    users_table_id = mdl.get_or_query_resource_id(ckan, "users.csv")
    traces_table_id = mdl.get_or_query_resource_id(ckan, "traces.csv")

    # datastore_search simple API
    cursor = ckan.datastore_search_cursor(users_table_id, limit=1)
    document = next(cursor)
    user_id = document["user_id"]
    cursor = ckan.datastore_search_cursor(traces_table_id, filters={"user_id": int(user_id)}, limit=10)
    for document in cursor:
        print(document)

    try:
        ckan.datastore_search_sql(f'SELECT * FROM "{users_table_id}"')
    except CkanSqlCapabilityError:
        print("datastore_search_sql is not accessible")

    query = f"""
    SELECT t.*, u.* FROM "{traces_table_id}" t
    JOIN "{users_table_id}" u ON t.user_id = u.user_id
    WHERE t.user_id = {user_id}
    LIMIT 10
    """
    cursor = ckan.datastore_search_sql_cursor(query, return_df=False)
    for document in cursor:
        print(document)

    print("Tests done.")


if __name__ == '__main__':
    ckan = CkanApi(None)
    ckan.initialize_from_cli_args()
    run(ckan)




#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements functions to initiate a DataStore.
"""
from typing import List, Tuple, Union, Collection
import os
from warnings import warn

import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.auxiliary.path import resolve_rel_path
from ckanapi_harvesters.builder.builder_errors import ResourceFileNotExistMessage
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element
from ckanapi_harvesters.auxiliary.ckan_model import UpsertChoice
from ckanapi_harvesters.builder.mapper_datastore_multi import RequestMapperABC
from ckanapi_harvesters.builder.builder_resource_datastore_multi_folder import BuilderDataStoreFolder
from ckanapi_harvesters.builder.builder_resource_multi_abc import FileChunkDataFrame


class BuilderDataStoreFile(BuilderDataStoreFolder):
    """
    Implementation supporting the reading of a file by chunks
    """
    def __init__(self, *, parent, name: str = None, format: str = None, description: str = None,
                 resource_id: str = None, download_url: str = None, file_name: str = None, options_string:str=None, base_dir:str=None):
        super().__init__(parent=parent, name=name, format=format, description=description, resource_id=resource_id,
                         download_url=download_url, dir_name="", options_string=options_string, base_dir=base_dir)
        self.file_size:int = 0
        self.upsert_method: UpsertChoice = UpsertChoice.Upsert
        self.file_name = file_name

    def copy(self, *, dest=None, parent=None):
        if dest is None:
            dest = BuilderDataStoreFile(parent=self.parent_package)
        super().copy(dest=dest, parent=parent)
        dest.file_name = self.file_name
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row, base_dir=base_dir)
        self.file_name: str = _string_from_element(row["file/url"], strip=True)

    @staticmethod
    def resource_mode_str() -> str:
        return "DataStore from File"

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return False

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = self.file_name
        return d

    # def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
    #     file_path = self.get_sample_file_path(resources_base_dir=resources_base_dir)
    #     if os.path.isfile(file_path):
    #         return None
    #     else:
    #         return ResourceFileNotExistMessage(self.name, ErrorLevel.Error, f"Missing file for resource {self.name}: {file_path}")

    ## upload chunks ----------------
    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        file_path = self.get_sample_file_path(resources_base_dir=resources_base_dir)
        if os.path.isfile(file_path):
            return None
        else:
            return ResourceFileNotExistMessage(self.name, ErrorLevel.Error, f"Missing file for resource {self.name}: {file_path}")

    def get_sample_file_path(self, resources_base_dir:str, ckan:Union[CkanApi,None]=None, file_index:int=0) -> str:
        return resolve_rel_path(resources_base_dir, self.file_name, field=f"File/URL of resource {self.name}")

    def list_local_files(self, resources_base_dir:str, ckan:CkanApi, cancel_if_present:bool=True) -> List[str]:
        file_path = self.get_sample_file_path(resources_base_dir=resources_base_dir)
        self.file_size = os.path.getsize(file_path)
        self.local_file_list = [file_path]
        self.local_file_size = [self.file_size]
        self.local_file_size_sum = self.file_size
        return self.local_file_list

    def get_local_file_offset(self, file_chunk: FileChunkDataFrame) -> int:
        return file_chunk.file_position

    def get_local_file_total_size(self) -> int:
        return self.file_size

    def get_local_file_count(self) -> int:
        return 1

    ## download ----------------
    def download_request(self, ckan: CkanApi, out_dir: str, *, full_download:bool=True,
                         force:bool=False, threads:int=1, return_data:bool=False) -> Union[pd.DataFrame,None]:
        if (not self.enable_download) and (not force):
            msg = f"Did not download resource {self.name} because download was disabled."
            warn(msg)
            return None
        if out_dir is not None:
            self.downloaded_destination = resolve_rel_path(out_dir, self.file_name, field=f"File/URL of resource {self.name}")
            if self.download_skip_existing and os.path.exists(self.downloaded_destination):
                return None
        resource_id = self.get_or_query_resource_id(ckan=ckan, error_not_found=self.download_error_not_found)
        if resource_id is None and not self.download_error_not_found:
            return None
        if self.local_file_format.append_allowed() and not return_data:
            download_generator = ckan.datastore_dump_page_generator(resource_id, search_all=full_download, progress_callback=self.progress_callback)
            first_df = None
            for df_download in download_generator:
                self.read_line_counter += len(df_download)
                df = self.df_mapper.df_download_alter(df_download, fields=self._get_fields_info())
                if out_dir is not None:
                    if first_df is None:
                        self.local_file_format.write_file(df, self.downloaded_destination, fields=self._get_fields_info())
                        first_df = df  # maybe the first DataFrame could be used in the append function to reproduce treatments
                    else:
                        self.local_file_format.append_file(df, self.downloaded_destination, fields=self._get_fields_info())
        else:
            df_download = ckan.datastore_dump(resource_id, search_all=full_download)
            self.read_line_counter += len(df_download)
            df = self.df_mapper.df_download_alter(df_download, fields=self._get_fields_info())
            if out_dir is not None:
                os.makedirs(out_dir, exist_ok=True)
                self.local_file_format.write_file(df, self.downloaded_destination, fields=self._get_fields_info())
            if return_data:
                return df
        return None

    def download_request_full(self, ckan: CkanApi, out_dir: str, threads:int=1, external_stop_event=None,
                              start_index:int=0, end_index:int=None, force:bool=False) -> None:
        return self.download_request(ckan=ckan, out_dir=out_dir, full_download=True, force=force)

    def to_builder_datastore_folder(self,
                                    *, dir_name:str=None, primary_key:List[str]=None,
                                    file_query_list:Collection[Tuple[str,dict]]=None) -> BuilderDataStoreFolder:
        resource_folder = BuilderDataStoreFolder(parent=self.parent_package)
        resource_folder._load_from_df_row(self._to_row())
        resource_folder.options_string = self.options_string
        resource_folder.data_cleaner_upload = self.data_cleaner_upload.copy()
        resource_folder.field_builders = self.field_builders
        if dir_name is not None:
            resource_folder.dir_name = dir_name
        elif isinstance(self, BuilderDataStoreFolder):
            resource_folder.dir_name = self.dir_name
        else:
            resource_folder.dir_name, _ = os.path.splitext(self.file_name)
        resource_folder.package_name = self.package_name
        if isinstance(self.df_mapper, RequestMapperABC):
            resource_folder.df_mapper = self.df_mapper.copy()
        else:
            resource_folder.df_mapper.df_upload_fun = self.df_mapper.df_upload_fun
            resource_folder.df_mapper.df_download_fun = self.df_mapper.df_download_fun
        if primary_key is not None or file_query_list is not None:
            resource_folder.setup_default_file_mapper(primary_key=primary_key, file_query_list=file_query_list)
        resource_folder.downloaded_file_query_list = file_query_list
        return resource_folder


#!python3
# -*- coding: utf-8 -*-
"""
Section of the package dedicated to the initialization of a CKAN package
"""

import os

# usage shortcuts
self_dir = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
from ..builder_package import example_package_xls, example_package_resources_dir

from . import builder_example
# from . import builder_example_aux_fun
from . import builder_example_generate_data
from . import builder_example_patch_upload
from . import builder_example_sample_dataset
from . import builder_example_tests_dev
from . import builder_example_test_sql
from . import builder_example_tests_offline
from . import builder_example_policy
from . import builder_example_download
from . import builder_example_tests




#!python3
# -*- coding: utf-8 -*-
"""
Adding support for geometries
"""
from typing import Any, Tuple, Union
from types import SimpleNamespace
import re
from warnings import warn

import numpy as np

from ckanapi_harvesters.auxiliary.lazy_imports import shapely, lazy_import_shapely
from ckanapi_harvesters.auxiliary.lazy_imports import pyproj, lazy_import_pyproj
from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.auxiliary.ckan_auxiliary import assert_or_raise
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_upload_1_basic import CkanDataCleanerUploadBasic
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_errors import UnexpectedGeometryError, FormatError, CleanerRequirementError

# mapping from Postgre geometric types to GeoJSON equivalents
# This does not enable the use of MultiPoint, MultiLine and MultiPolygon
postgre_geojson_mapping = {
    "point": "Point",
    "path": "LineString",
    "polygon": "Polygon",
   }

multi_geometry_types = {'MultiPoint', 'MultiLineString', 'MultiPolygon', 'GeometryCollection'}

def shapely_geometry_from_value(value:Any) -> Union[shapely.Geometry,None]:
    global shapely
    shapely = lazy_import_shapely()
    if shapely.Geometry is None:
        raise CleanerRequirementError("shapely", "geometry")
    if value is None:
        return None
    elif isinstance(value, shapely.Geometry):
        return value
    elif isinstance(value, str):
        if len(value) == 0:
            return None
        elif value[0] in {'{', '[', '('}:
            return shapely.from_geojson(value)
        elif re.match("[a-zA-Z]+\\(.+\\)", value):
            return shapely.from_wkt(value)
        elif re.match("[0-9A-F]+", value):
            return shapely.from_wkb(value)
        else:
            raise FormatError(value, "geometry")
    elif isinstance(value, dict):
        return shapely.geometry.shape(value)
    else:
        raise FormatError(value, "geometry")

def has_invalid_coordinates(shape: shapely.Geometry) -> Tuple[bool, bool]:
    all_nonfinite = True
    any_nonfinite = False
    if shape.geom_type in multi_geometry_types:
        # recurse
        for sub_geom in shape.geoms:
            sub_any_nonfinite, sub_all_nonfinite = has_invalid_coordinates(sub_geom)
            any_nonfinite = any_nonfinite or sub_any_nonfinite
            all_nonfinite = all_nonfinite and sub_all_nonfinite
    else:
        for coord in shape.coords:
            if any(not (np.isfinite(c)) for c in coord):
                any_nonfinite = True
            else:
                all_nonfinite = False
    return any_nonfinite, all_nonfinite


class CkanDataCleanerUploadGeom(CkanDataCleanerUploadBasic):
    def __init__(self):
        super().__init__()

    @staticmethod
    def get_class_keyword() -> str:
        return "GeoJSON"

    def _replace_standard_value_bypass(self, value: Any, field: CkanField, *, field_data_type: str) -> Tuple[Any, bool]:
        if value is None:
            return value, False
        elif field_data_type == "geometry" or field_data_type.startswith("geometry("):  #  and field.internal_attrs.geometry_as_source:
            value_shape = shapely_geometry_from_value(value)
            any_nonfinite, all_nonfinite = has_invalid_coordinates(value_shape)
            if any_nonfinite:
                if not all_nonfinite:
                    msg = "Shape was ignored because it contains non-finite coordinates: " + shapely.to_geojson(value)
                    warn(msg)
                return None, True
            field_geojson_type = field.internal_attrs.geometry_type
            if field_geojson_type is not None and not field_geojson_type.casefold() == "geometry":
                assert_or_raise(value_shape.geom_type.casefold() == field_geojson_type.casefold(), UnexpectedGeometryError(value_shape.geom_type, field_geojson_type))
            if field.internal_attrs.epsg_source is not None and field.internal_attrs.epsg_target is not None:
                if not field.internal_attrs.epsg_source == field.internal_attrs.epsg_target:
                    global pyproj
                    pyproj = lazy_import_pyproj()
                    if pyproj is None:
                        raise CleanerRequirementError("pyproj", "geometry projection")
                    crs_source = pyproj.CRS.from_epsg(field.internal_attrs.epsg_source)
                    crs_target = pyproj.CRS.from_epsg(field.internal_attrs.epsg_target)
                    transformer = pyproj.Transformer.from_crs(crs_source, crs_target, always_xy=True)
                    value_shape_prev = value_shape  # for debug
                    value_shape = shapely.transform(value_shape, transformer.transform, interleaved=False)
            value_wkb = shapely.to_wkb(value_shape, hex=True)
            return value_wkb, True
        elif field_data_type in postgre_geojson_mapping.keys():
            if field.internal_attrs.geometry_as_source:
                value_shape = shapely_geometry_from_value(value)
                geojson_type = postgre_geojson_mapping[field_data_type]
                assert_or_raise(value_shape.geom_type == geojson_type, UnexpectedGeometryError(value_shape.geom_type, geojson_type))
                coordinates = shapely.get_coordinates(value_shape)
                if field_data_type == "point":
                    # representation: (x,y)
                    return str(tuple(coordinates)), True
                elif field_data_type == "path":
                    # representation: [(x1,y1),...]
                    return str([tuple(point) for point in coordinates]), True
                elif field_data_type == "polygon":
                    # representation: ((x1,y1),...)
                    return str(tuple([tuple(point) for point in coordinates])), True
                else:
                    raise NotImplementedError()
            else:
                return str(value), True
        else:
            return super()._replace_standard_value_bypass(value, field, field_data_type=field_data_type)




#!python3
# -*- coding: utf-8 -*-
"""
Class to parameterize and establish an SSH tunnel to a distant server
"""
from typing import Union
import io
import argparse
import getpass

from ckanapi_harvesters.auxiliary.lazy_imports import SSHTunnelForwarder, lazy_import_ssh_tunnel_SSHTunnelForwarder
from ckanapi_harvesters.auxiliary.path import path_rel_to_dir
from ckanapi_harvesters.auxiliary.ckan_errors import RequirementError
from ckanapi_harvesters.auxiliary.login import Login
from ckanapi_harvesters.auxiliary.proxy_config import ProxyConfig


class SshLogin(Login):
    LOGIN_FILE_ENVIRON = "SSH_AUTH_FILE"

    @staticmethod
    def _setup_cli_parser(parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        if parser is None:
            parser = argparse.ArgumentParser(description="SSH login credentials initialization", add_help=False)
        parser.add_argument("--ssh-login-file", type=str,
                            help="Path to a text file containing SSH login credentials for authentification (user, password)")
        parser.add_argument("--ssh-user", type=str,
                            help="SSH user name (prefer using a file)")
        parser.add_argument("--ssh-password", type=str,
                            help="SSH user password (prefer using a file)")
        return parser

    def _cli_args_apply(self, args: argparse.Namespace, *, base_dir: str = None, error_not_found: bool = True) -> None:
        if args.login_file is not None:
            self.load_from_file(args.ssh_login_file, base_dir=base_dir, error_not_found=error_not_found)
        if args.ssh_user is not None:
            self.username = args.ssh_user
        if args.ssh_password is not None:
            self.password = args.ssh_password

    def input(self):
        """
        Prompt the user to input the login credentials in the console window.

        :return:
        """
        value = input("Please enter SSH user name: ")
        self.username = value
        value = getpass.getpass("Please enter SSH password: ")
        self.password = value


class SshTunnel:
    def __init__(self, *, remote_host: str = None, remote_port: int = None,
                 ssh_host: str=None, ssh_port: int=None, ssh_login: SshLogin=None,
                 ssh_login_file: str=None, ssh_pkey_file: str=None,
                 proxy:ProxyConfig=None) -> None:
        """
        SSH Tunnel parameterization functions.

        SSH remote is to be configured by the caller. The other attributes can be configured by the CLI.

        :param remote_host: Remote bind host. This is the service which is not exposed in clear, on server side.
        :param remote_port: Remote bind port.
        :param ssh_host: Remote SSH server host.
        :param ssh_port: Remote SSH server port.
        :param ssh_login_file: Login to connect to the SSH server.
        :param ssh_pkey_file: Path to the SSH private key file.
        """
        if ssh_login is None:
            ssh_login = SshLogin()
        if proxy is None:
            proxy = ProxyConfig()
        self.remote_host: str = remote_host
        self.remote_port: int = remote_port
        self.ssh_host: str = ssh_host
        self.ssh_port: int = ssh_port
        self.ssh_pkey_file = ssh_pkey_file
        self.ssh_login: SshLogin = ssh_login
        self.socks_proxy: ProxyConfig = proxy
        self.server: Union[SSHTunnelForwarder, None] = None
        if ssh_login_file is not None:
            self.ssh_login.load_from_file(ssh_login_file)
        self.verbose: bool = False
        raise NotImplementedError("Not tested")

    def __str__(self):
        if self.server is None:
            return f"SSH Tunnel to {self.remote_host}:{self.remote_port} (disconnected)"
        else:
            return f"SSH Tunnel to {self.remote_host}:{self.remote_port} ({self.get_tunnel_url()})"

    def __copy__(self):
        return self.copy()

    def copy(self) -> "SshTunnel":
        dest = SshTunnel()
        dest.remote_host = self.remote_host
        dest.remote_port = self.remote_port
        dest.ssh_host = self.ssh_host
        dest.ssh_port = self.ssh_port
        dest.ssh_login = self.ssh_login.copy()
        dest.socks_proxy = self.socks_proxy.copy()
        dest.ssh_pkey_file = self.ssh_pkey_file
        dest.verbose = self.verbose
        return dest

    @staticmethod
    def _setup_cli_parser(parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        if parser is None:
            parser = argparse.ArgumentParser(description="SSH tunnel initialization", add_help=False)
        SshLogin._setup_cli_parser(parser)
        parser.add_argument("--ssh-host", type=str,
                            help="Remote SSH server host")
        parser.add_argument("--ssh-port", type=int,
                            help="Remote SSH server port")
        parser.add_argument("--ssh-key-file", type=str,
                            help="Path to the private key file")
        return parser

    def print_help_cli(self, display:bool=True) -> str:
        parser = self._setup_cli_parser()
        if display:
            parser.print_help()
        with io.StringIO() as stream:
            parser.print_help(stream)
            return stream.getvalue()

    def _cli_args_apply(self, args: argparse.Namespace, *, base_dir: str = None, error_not_found: bool = True) -> None:
        if args.ssh_host is not None:
            self.ssh_host = args.ssh_host
        if args.ssh_port is not None:
            self.ssh_port = args.ssh_port
        if args.ssh_key_file is not None:
            self.ssh_pkey_file = path_rel_to_dir(args.ssh_key_file, base_dir=base_dir)

    def start_tunnel(self):
        # lazy import: only if necessary here
        global SSHTunnelForwarder
        SSHTunnelForwarder = lazy_import_ssh_tunnel_SSHTunnelForwarder()
        # ----------
        if SSHTunnelForwarder is None:
            raise RequirementError("sshtunnel", "SshTunnel")
        ssh_proxy = self.socks_proxy.get_host_port() if self.socks_proxy.is_defined() else None
        self.server = SSHTunnelForwarder(
            (self.ssh_host, self.ssh_port),
            ssh_username=self.ssh_login.username,
            ssh_password=self.ssh_login.password,
            ssh_pkey=self.ssh_pkey_file,
            remote_bind_address=(self.remote_host, self.remote_port),
            ssh_proxy=ssh_proxy,
        )
        self.server.start()
        print(f"SSH tunnel established: localhost:{self.server.local_bind_port} -> {self.remote_host}:{self.remote_port}")

    def get_tunnel_host(self) -> str:
        if self.server is None:
            raise ConnectionError("SSH tunnel was not opened")
        return "localhost"
    def get_tunnel_port(self) -> int:
        if self.server is None:
            raise ConnectionError("SSH tunnel was not opened")
        return self.server.local_bind_port
    def get_tunnel_url(self) -> str:
        if self.server is None:
            raise ConnectionError("SSH tunnel was not opened")
        return f"localhost:{self.server.local_bind_port}"

    def is_connected(self) -> bool:
        return self.server is not None

    def is_defined(self) -> bool:
        return self.ssh_host is not None

    def close_tunnel(self):
        """
        Close SSH tunnel. Please close underlying connections before.
        """
        if self.server is not None:
            self.server.stop()
            self.server = None

    def __del__(self):
        self.ssh_login.__del__()
        self.close_tunnel()

if __name__ == '__main__':
    tunnel = SshTunnel(remote_host="www.google.com", remote_port=443,
                       ssh_host="", ssh_port=22,  # TODO: define host
                       ssh_login=SshLogin("demo", "password"))
    print("SSH tunnel CLI-format options:")
    tunnel.print_help_cli()

    print("Test connection")
    import requests
    tunnel.start_tunnel()
    A = requests.get(f"{tunnel.remote_host}:{tunnel.remote_port}")
    B = requests.get(tunnel.get_tunnel_url())
    tunnel.close_tunnel()
    print("Finished")




#!python3
# -*- coding: utf-8 -*-
"""
Basic parameters for the CkanApi class
"""
from typing import Union, Tuple
import copy
from warnings import warn
import argparse

import requests
from requests.auth import AuthBase

from ckanapi_harvesters.auxiliary.proxy_config import ProxyConfig
from ckanapi_harvesters.auxiliary.ckan_auxiliary import CkanIdFieldTreatment
from ckanapi_harvesters.auxiliary.ckan_configuration import allow_no_ca
from ckanapi_harvesters.auxiliary.ckan_errors import NoCAVerificationError
from ckanapi_harvesters.auxiliary.path import path_rel_to_dir

default_df_download_id_field_treatment: CkanIdFieldTreatment = CkanIdFieldTreatment.SetIndex


class CkanApiParamsBasic:
    def __init__(self, *, proxies:Union[str,dict,ProxyConfig]=None,
                 ckan_headers:dict=None, http_headers:dict=None):
        """
        :param proxies: proxies to use for requests
        :param ckan_headers: headers to use for requests, only to the CKAN server
        :param http_headers: headers to use for requests, for all requests, including external requests and to the CKAN server
        """
        if ckan_headers is None: ckan_headers = {}
        if http_headers is None: http_headers = {}
        # HTTP parameters
        self._proxy_config: ProxyConfig = ProxyConfig.from_str_or_config(proxies)
        self.user_agent: Union[str,None] = None
        self.http_headers: dict = http_headers
        self.ckan_headers: dict = ckan_headers
        self._ckan_ca: Union[bool, str, None] = True  # use to specify a path to a custom CA certificate for the CKAN server (see also environment variable REQUESTS_CA_BUNDLE)
        self._extern_ca: Union[bool, str, None] = True  # use to specify a path to a custom CA certificate for external connexions (see also environment variable REQUESTS_CA_BUNDLE)
        # debug parameters
        self.store_last_response: bool = True
        self.store_last_response_debug_info: bool = True
        # modes
        self.dry_run: bool = False  # if True, no requests are sent to the server (for debugging purposes)
        # limits
        self.default_limit_list:Union[int,None] = 100   # limit the number of entries per list response (used as default value)
        self.default_limit_read:Union[int,None] = 5000  # limit the number of entries per response (used as default value)
        self.max_requests_attempts:int = 3  # when performing an API action, retry depending on the received HTTP error code
        self.action_requests_retry_always:bool = False  # when an error occurred during an API action request, systematically retry
        self.max_requests_count:int = 0  # when automating multiple requests, the number of requests is limited by this parameter. 0 means no limit.
        # timeouts
        self.time_between_attempts:float = 11  # when a retry was decided, wait this time before a new tentative (* request number)
        self.response_time_wait_threshold:Union[None,float] = 30  # threshold to determine a long response time => wait the same duration in return
        self.multi_requests_timeout:float = 0  # when automating multiple requests, the total time elapsed is limited by this parameter (evaluated between each request). 0 means no limit.
        self.multi_requests_time_between_requests:float = 0.100  # when automating multiple requests, wait this additional time (in seconds) between each request
        self.requests_timeout:Union[float,None] = 100  # timeout per request sent to the requests module
        # verbosity
        self.verbose_multi_requests:bool = False
        self.verbose_request:bool = False
        self.verbose_request_error:bool = True
        self.verbose_extra:bool = True

    def copy(self, *, dest=None):
        if dest is None:
            dest = CkanApiParamsBasic()
        dest._proxy_config = self._proxy_config.copy()
        dest.user_agent = copy.deepcopy(self.user_agent)
        dest.http_headers = copy.deepcopy(self.http_headers)
        dest.ckan_headers = copy.deepcopy(self.ckan_headers)
        dest._ckan_ca = self._ckan_ca
        dest._extern_ca = self._extern_ca
        dest.dry_run = self.dry_run
        dest.store_last_response = self.store_last_response
        dest.store_last_response_debug_info = self.store_last_response_debug_info
        dest.default_limit_list = self.default_limit_list
        dest.default_limit_read = self.default_limit_read
        dest.max_requests_count = self.max_requests_count
        dest.multi_requests_timeout = self.multi_requests_timeout
        dest.multi_requests_time_between_requests = self.multi_requests_time_between_requests
        dest.requests_timeout = self.requests_timeout
        dest.verbose_multi_requests = self.verbose_multi_requests
        dest.verbose_request = self.verbose_request
        dest.verbose_request_error = self.verbose_request_error
        dest.verbose_extra = self.verbose_extra
        return dest

    @property
    def proxies(self) -> dict:
        return self._proxy_config.proxies
    @proxies.setter
    def proxies(self, proxies:dict) -> None:
        self._proxy_config.proxies = proxies
    @property
    def proxy_string(self) -> str:
        return self._proxy_config.proxy_string
    @proxy_string.setter
    def proxy_string(self, proxies:str) -> None:
        self._proxy_config.proxy_string = proxies
    @property
    def proxy_auth(self) -> Union[AuthBase, Tuple[str,str]]:
        return self._proxy_config.proxy_auth
    @proxy_auth.setter
    def proxy_auth(self, proxy_auth:Union[AuthBase, Tuple[str,str]]) -> None:
        self._proxy_config.proxy_auth = proxy_auth
    @property
    def ckan_ca(self) -> Union[bool,str,None]:
        return self._ckan_ca
    @ckan_ca.setter
    def ckan_ca(self, ca_cert:Union[bool,str,None]) -> None:
        if ca_cert is not None and isinstance(ca_cert, bool) and not ca_cert:
            if not allow_no_ca:
                raise NoCAVerificationError()
            else:
                msg = "CA verification has been disabled. Only allow in a local environment!"
                warn(msg)
        self._ckan_ca = ca_cert
    @property
    def extern_ca(self) -> Union[bool,str,None]:
        return self._extern_ca
    @extern_ca.setter
    def extern_ca(self, ca_cert:Union[bool,str,None]) -> None:
        if ca_cert is not None and isinstance(ca_cert, bool) and not ca_cert:
            if not allow_no_ca:
                raise NoCAVerificationError()
            else:
                msg = "CA verification has been disabled. Only allow in a local environment!"
                warn(msg)
        self._extern_ca = ca_cert

    @staticmethod
    def _setup_cli_ckan_parser__params(parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        """
        Define or add CLI arguments to initialize a CKAN API connection
        parser help message:

        CKAN API connection parameters initialization

        :param parser: option to provide an existing parser to add the specific fields needed to initialize a CKAN API connection
        :return:
        """
        if parser is None:
            parser = argparse.ArgumentParser(description="CKAN API connection parameters initialization", add_help=False)
        ProxyConfig._setup_cli_proxy_parser(parser)  # add arguments --proxy --http-proxy --https-proxy --no-proxy --proxy-auth-file
        parser.add_argument("--ckan-ca", type=str,
                            help="CKAN CA certificate location (.pem file)")
        parser.add_argument("--extern-ca", type=str,
                            help="CA certificate location for extern connexions (.pem file)")
        parser.add_argument("--user-agent", type=str,
                            help="User agent for HTTP requests")
        parser.add_argument("-l", "--limit", type=int,
                            help="Number of rows per request (upload/download)")
        parser.add_argument("-v", "--verbose",
                            help="Option to set verbosity", action="store_true", default=False)
        parser.add_argument("--time-between-requests", type=float,
                            help="Time between upload/download requests (seconds) - recommended: 0.1 seconds")
        # parser.add_argument("--external-code", action="store_true",
        #                     help="Enable external code execution for builder (only enable for trusted sources)")
        return parser

    def _cli_ckan_args_apply(self, args: argparse.Namespace, *, base_dir:str=None, error_not_found:bool=True,
                             default_proxies:dict=None, proxy_headers:dict=None) -> None:
        """
        Apply the arguments parsed by the argument parser defined by _setup_cli_ckan_parser

        :param args:
        :param base_dir: base directory to find the CKAN API key file, if a relative path is provided
            (recommended: leave None to use cwd)
        :param error_not_found: option to raise an exception if the CKAN API key file is not found
        :param default_proxies: proxies used if proxies="default"
        :param proxy_headers: headers used to access the proxies, generally for authentication
        :return:
        """
        proxy_config = ProxyConfig.from_cli_args(args, base_dir=base_dir, error_not_found=error_not_found,
                                                 default_proxies=default_proxies, proxy_headers=proxy_headers)
        if proxy_config is not None:
            self._proxy_config = proxy_config
        if args.ckan_ca is not None:
            self.ckan_ca = path_rel_to_dir(args.ckan_ca, base_dir=base_dir)
        if args.extern_ca is not None:
            self.extern_ca = path_rel_to_dir(args.extern_ca, base_dir=base_dir)
        if args.user_agent is not None:
            self.user_agent = args.user_agent
        if args.time_between_requests is not None:
            self.multi_requests_time_between_requests = args.time_between_requests
        if args.limit is not None:
            self.default_limit_read = args.limit
        # if args.verbose is not None:
        #     self.set_verbosity(args.verbose)
        # if args.external_code:
        #     unlock_external_code_execution()
        print(args)

class CkanApiDebug:
    def __init__(self):
        self.ckan_request_counter: int = 0
        self.extern_request_counter: int = 0
        self.last_response: Union[requests.Response, None] = None  # field containing the last response, for debug purposes
        self.last_response_request_count: int = 0
        self.multi_requests_last_successful_offset: int = 0  # last used offset when multiple queries are performed. This can be used in order to restart an update/download sequence in case of an error.




#!python3
# -*- coding: utf-8 -*-
"""
Functions to clean data before upload.
"""
from typing import Union, List, Any, Dict, Set, Tuple
from abc import ABC, abstractmethod
from collections import OrderedDict

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.auxiliary.ckan_auxiliary import assert_or_raise

non_finite_authorized_types = {"numeric", "float4", "float8", "float2"}
real_number_types = non_finite_authorized_types
dtype_mapper = {
    "float64": "numeric",
    "int64": "numeric",
    "datetime64[ns]": "timestamp",
}


class CkanDataCleanerABC(ABC):
    """
    Data cleaner abstract base class.

    A table is defined by a list of fields with a data type.
    Each row can specify the value of all/some fields.
    When a value is nested (dictionary or list), the functions iterate over the values of these elements with a recursive implementation.
    These elements are called sub-values.
    """
    def __init__(self):
        # options
        self.param_enable:bool = True  # global activation flag
        self.param_replace_forbidden:bool = False  # option to replace all other forbidden values (Infs) by None
        self.param_cast_types:bool = True  # option to cast to strings fields which have text data type
        self.param_apply_field_subs:bool = True  # option to apply suggested field renamings (True by default because these are suggested only when necessary)
        self.param_apply_field_changes:bool = False  # option to apply suggested field type changes
        self.param_raise_error:bool = False  # recommended: do not raise an error: the CKAN server will
        self.param_create_new_fields:bool = True  # option to enable the requests to create missing fields in the CKAN DataStore (this requires the specific function to be called)
        self.param_verbose:bool = True
        self.param_field_subs:Dict[str,str] = {}  # user-imposed field name substitutions
        self.param_field_primary_key:Union[List[str],None] = None
        # outputs
        self.fields_encountered:OrderedDict[str,None] = OrderedDict()
        self.warnings:Dict[str,Set[str]] = {}
        self.fields_new:OrderedDict[str,CkanField] = OrderedDict()
        self.field_changes:Dict[str,CkanField] = {}
        self.field_subs:Dict[str, str] = {}
        self.field_subs_path:Dict[str, str] = {}
        self.field_suggested_primary_key:Union[List[str],None] = None
        self.field_suggested_index:Set[str] = set()
        self._new_columns_in_row: Dict[str,Any] = None  # is initialized at each row

    @staticmethod
    @abstractmethod
    def get_class_keyword() -> str:
        """
        Returns the name of the class, according to data_cleaner_dict defined in data_cleaner_init.py.
        This name is used to setup the data cleaner for a resource builder.
        """
        raise NotImplementedError()

    def clear_outputs_new_dataframe(self):
        self.fields_encountered = OrderedDict()
        self.warnings = {}
        self.fields_new = OrderedDict()
        self.field_changes = {}
        self.field_subs = {}
        self.field_subs_path = {}
        self.field_suggested_primary_key = self.param_field_primary_key
        self.field_suggested_index = set()
        self._new_columns_in_row = None

    def clear_all_outputs(self):
        """
        Some values must not be cleared for each DataFrame upload.
        The cleaner is stateful for certain values cleared only here.
        """
        self.clear_outputs_new_dataframe()

    @abstractmethod
    def copy(self, dest=None):
        dest.param_enable = self.param_enable
        dest.param_replace_forbidden = self.param_replace_forbidden
        dest.param_apply_field_subs = self.param_apply_field_subs
        dest.param_apply_field_changes = self.param_apply_field_changes
        dest.param_raise_error = self.param_raise_error
        dest.param_create_new_fields = self.param_create_new_fields
        dest.param_verbose = self.param_verbose
        dest.clear_outputs_new_dataframe()
        return dest

    def __copy__(self):
        return self.copy()

    ## Field type detection ------------------
    def _detect_standard_field_bypass(self, field_name: str, values: Union[Any, pd.Series]) -> Union[CkanField,None]:
        """
        Auxiliary function of create_new_field to detect field type used to bypass the default criteria.
        """
        return None

    def _detect_non_standard_field(self, field_name: str, values: Union[Any, pd.Series]) -> CkanField:
        """
        Auxiliary function of create_new_field to detect field type used if the default criteria did not match any specific case.
        """
        return CkanField(field_name, "text")

    @abstractmethod
    def create_new_field(self, field_name:str, values: Union[Any, pd.Series]) -> CkanField:
        """
        This method adds a new field definition
        """
        raise NotImplementedError()

    @abstractmethod
    def detect_field_types_and_subs(self, records: Union[List[dict], pd.DataFrame],
                                    known_fields:OrderedDict[str, CkanField]=None) -> OrderedDict[str, str]:
        """
        This function detects the initial fields and necessary field renamings
        """
        raise NotImplementedError()

    ## Records cleansing -------------
    @abstractmethod
    def clean_value_field(self, value: Any, field: CkanField) -> Any:
        """
        Cleaning of a value. A value is directly the value of a cell.
        """
        raise NotImplementedError()

    def _replace_standard_value_bypass(self, value: Any, field: CkanField, *, field_data_type: str) -> Tuple[Any, bool]:
        """
        Auxiliary function of clean_value_field to perform type castings/checks used to bypass the default criteria.
        """
        return None, False

    def _replace_non_standard_value(self, value: Any, field: CkanField, *, field_data_type: str) -> Any:
        """
        Auxiliary function of clean_value_field to perform type castings/checks used if none of the default criteria were met.
        """
        return value

    @abstractmethod
    def _clean_subvalue(self, subvalue: Any, field: CkanField, path: str, level: int,
                                   *, field_data_type: str) -> Any:
        """
        Cleaning of a subvalue. A subvalue is a value within a nested cell.
        """
        raise NotImplementedError()

    def _replace_standard_subvalue_bypass(self, subvalue:Any, field:CkanField, path:str, level:int,
                                          *, field_data_type:str) -> Tuple[Any,bool]:
        """
        Auxiliary function of _clean_subvalue to perform type castings/checks used to bypass the default criteria.
        """
        return None, False

    def _replace_non_standard_subvalue(self, subvalue:Any, field:CkanField, path:str, level:int,
                                       *, field_data_type:str) -> Any:
        """
        Auxiliary function of _clean_subvalue to perform type castings/checks used if none of the default criteria were met.
        """
        return subvalue

    def _add_field_from_path(self, path:str, data_type:str, new_field_name:str=None,
                             suggest_index:bool=True, notes:str=None) -> None:
        """
        Auxiliary method to define a new column from a nested object.
        """
        if new_field_name is None:
            new_field_name = path.replace(".", "_")
        assert_or_raise(new_field_name not in self.fields_encountered, KeyError(f"{new_field_name} already exists and cannot be replaced"))
        self.fields_new[new_field_name] = CkanField(new_field_name, data_type, notes=notes)
        self.field_subs_path[path] = new_field_name
        if suggest_index:
            self.field_suggested_index.add(new_field_name)
        self.fields_encountered[new_field_name] = None

    @abstractmethod
    def clean_records(self, records: Union[List[dict], pd.DataFrame], known_fields:Union[OrderedDict[str, CkanField], None],
                      *, inplace:bool=False) -> Union[List[dict], pd.DataFrame]:
        """
        Main function to clean a list of records.

        :param records:
        :param known_fields:
        :param inplace:
        :return:
        """
        raise NotImplementedError()

    @abstractmethod
    def _clean_final_steps(self, records: Union[List[dict], pd.DataFrame], fields:Union[OrderedDict[str, CkanField], None],
                           known_fields:Union[OrderedDict[str, CkanField], None]) -> Union[List[dict], pd.DataFrame]:
        """
        Method called at the end of clean_records
        """
        raise NotImplementedError()

    def _extra_checks(self, records: Union[List[dict], pd.DataFrame], fields:Union[OrderedDict[str, CkanField], None]) -> None:
        """
        Method called at the end of _clean_final_steps
        """
        pass

    ### post-treatments -------------
    def apply_new_fields_request(self, ckan, resource_id:str):
        """
        This method performs the field patch if a new field was detected.
        Call before upsert.
        """
        if self.param_create_new_fields and len(self.fields_new) > 0:
            ckan.datastore_field_patch(resource_id, fields_update=self.fields_new)

    def merge_field_changes(self, fields:List[dict]=None) -> List[dict]:
        """
        This method merges the fields argument of a datastore_create with the fields detected by the data cleaner.
        Fields already defined in the fields argument are not overwritten.
        """
        if fields is not None:
            if isinstance(fields, dict):
                fields_dict = fields
            else:
                fields_dict = OrderedDict([(field_dict["id"], CkanField.from_ckan_dict(field_dict))  for field_dict in fields])
        else:
            fields_dict = OrderedDict()
        if len(self.fields_new) > 0:
            for field_name, field_info in self.fields_new.items():
                if field_name not in fields_dict.keys():
                    fields_dict[field_name] = field_info
                else:
                    # was not new?  => merge changes?
                    fields_dict[field_name] = fields_dict[field_name].merge(field_info)
                    raise RuntimeError()
            # fields_dict = dict_recursive_update(fields_dict, {field_info.name: field_info.to_ckan_dict() for field_info in self.fields_new.values()})
        if self.param_apply_field_changes:
            if len(self.field_changes) > 0:
                for field_name, field_info in self.field_changes.items():
                    if field_name not in fields_dict.keys():
                        # new?  => create?
                        fields_dict[field_name] = field_info
                        raise RuntimeError()
                    else:
                        fields_dict[field_name] = fields_dict[field_name].merge(field_info)
                # fields_dict = dict_recursive_update(fields_dict, {field_info.name: field_info.to_ckan_dict() for field_info in self.field_changes.values()})
            return [field_info.to_ckan_dict() for field_info in fields_dict.values()]
        else:
            return fields


class DataCleanerNone(CkanDataCleanerABC):
    """
    Implementation which does nothing. Placeholder to explicitly mention a data cleaner must not be used.
    """
    @staticmethod
    def get_class_keyword() -> str:
        return "None"

    def copy(self, dest=None) -> "DataCleanerNone":
        if dest is None:
            dest = DataCleanerNone()
        super().copy(dest=dest)
        return dest

    def clean_records(self, records: Union[List[dict], pd.DataFrame], known_fields:Union[OrderedDict[str, CkanField], None],
                      *, inplace:bool=False) -> Union[List[dict], pd.DataFrame]:
        return records

    def _clean_final_steps(self, records: Union[List[dict], pd.DataFrame], fields:Union[OrderedDict[str, CkanField], None],
                           known_fields:Union[OrderedDict[str, CkanField], None]) -> Union[List[dict], pd.DataFrame]:
        return records

    def clean_value_field(self, value: Any, field: CkanField) -> Any:
        return value

    def _clean_subvalue(self, subvalue: Any, field: CkanField, path: str, level: int,
                                   *, field_data_type: str) -> Any:
        return subvalue

    def create_new_field(self, field_name:str, values: Union[Any, pd.Series]) -> CkanField:
        if field_name in self.fields_new.keys():
            return self.fields_new[field_name]
        else:
            raise KeyError(f"Field {field_name} does not exist")

    def detect_field_types_and_subs(self, records: Union[List[dict], pd.DataFrame],
                                    known_fields:OrderedDict[str, CkanField]=None) -> OrderedDict[str, str]:
        raise NotImplementedError()




#!python3
# -*- coding: utf-8 -*-
"""
Data format policy representation and enforcing
"""
from collections import OrderedDict
from typing import Union

from ckanapi_harvesters.auxiliary.ckan_auxiliary import empty_str_to_None

class DataFormatPolicyOutputCustomFields:
    default_report_timestamp_field: Union[str, None] = "Last Report Timestamp"
    default_score_custom_field: Union[str, None] = "Data Policy Score"
    default_report_custom_field: Union[str, None] = "Data Policy Report"
    default_package_filestore_size_custom_field: Union[str, None] = "FileStore Size"
    default_package_external_size_custom_field: Union[str, None] = "External Size"
    default_package_datastore_size_custom_field: Union[str, None] = "DataStore Size"
    default_package_datastore_rowcount_custom_field: Union[str, None] = "DataStore Rows"

    def __init__(self):
        self.package_score_field:Union[str,None] = DataFormatPolicyOutputCustomFields.default_score_custom_field
        self.package_report_field:Union[str,None] = DataFormatPolicyOutputCustomFields.default_report_custom_field
        # configured for admin report:
        self.report_timestamp_field:Union[str,None] = DataFormatPolicyOutputCustomFields.default_report_timestamp_field
        self.package_filestore_size_field:Union[str,None] = DataFormatPolicyOutputCustomFields.default_package_filestore_size_custom_field
        self.package_external_size_field:Union[str,None] = DataFormatPolicyOutputCustomFields.default_package_external_size_custom_field
        self.package_datastore_size_field:Union[str,None] = DataFormatPolicyOutputCustomFields.default_package_datastore_size_custom_field
        self.package_datastore_rowcount_field:Union[str,None] = DataFormatPolicyOutputCustomFields.default_package_datastore_rowcount_custom_field

    def copy(self) -> "DataFormatPolicyOutputCustomFields":
        dest = DataFormatPolicyOutputCustomFields()
        dest.package_score_field = self.package_score_field
        dest.package_report_field = self.package_report_field
        dest.report_timestamp_field = self.report_timestamp_field
        dest.package_filestore_size_field = self.package_filestore_size_field
        dest.package_external_size_field = self.package_external_size_field
        dest.package_datastore_size_field = self.package_datastore_size_field
        dest.package_datastore_rowcount_field = self.package_datastore_rowcount_field
        return dest

    def to_dict(self) -> dict:
        output_custom_field_config = OrderedDict()
        output_custom_field_config["report_timestamp"] = self.report_timestamp_field if self.report_timestamp_field else ""
        output_custom_field_config["package_score"] = self.package_score_field if self.package_score_field else ""
        output_custom_field_config["package_report"] = self.package_report_field if self.package_report_field else ""
        output_custom_field_config["package_filestore_size"] = self.package_filestore_size_field if self.package_filestore_size_field else ""
        output_custom_field_config["package_external_size"] = self.package_external_size_field if self.package_external_size_field else ""
        output_custom_field_config["package_datastore_size"] = self.package_datastore_size_field if self.package_datastore_size_field else ""
        output_custom_field_config["package_datastore_rowcount"] = self.package_datastore_rowcount_field if self.package_datastore_rowcount_field else ""
        return output_custom_field_config

    def _load_from_dict(self, d:dict) -> None:
        if d is None:
            return
        self.report_timestamp_field = empty_str_to_None(d.get("report_timestamp"))
        self.package_score_field = empty_str_to_None(d.get("package_score"))
        self.package_report_field = empty_str_to_None(d.get("package_report"))
        self.package_filestore_size_field = empty_str_to_None(d.get("package_filestore_size"))
        self.package_external_size_field = empty_str_to_None(d.get("package_external_size"))
        self.package_datastore_size_field = empty_str_to_None(d.get("package_datastore_size"))
        self.package_datastore_rowcount_field = empty_str_to_None(d.get("package_datastore_rowcount"))

    def set_metadata_policy_fields(self, *, score: str = None, report: str = None):
        if score is not None:
            self.package_score_field = empty_str_to_None(score)
        if report is not None:
            self.package_report_field = empty_str_to_None(report)

    def disable_all(self):
        self.package_score_field = None
        self.package_report_field = None
        self.report_timestamp_field = None
        self.package_filestore_size_field = None
        self.package_external_size_field = None
        self.package_datastore_size_field = None
        self.package_datastore_rowcount_field = None




#!python3
# -*- coding: utf-8 -*-
"""
Parameters which apply to the package
"""
from typing import Union


## Package containing CKAN data format policy
configuration_package_name = "configuration"
policy_resource = "data_format_policy.json"
datastore_sample_resource = "sample_datastore.csv"
datastore_default_upload_index_col_name: Union[str,None] = "py_upload_index"
datastore_default_source_file_col_name: Union[str,None] = "py_source_file"
module_fields = {datastore_default_upload_index_col_name, datastore_default_source_file_col_name}

default_ckan_has_postgis: bool = True
epsg_wgs84 = 4326  # WGS-84
default_ckan_target_epsg: Union[int,None] = epsg_wgs84  # default target geodesic system


allow_no_ca = False

def unlock_no_ca(value:bool=True) -> None:
    """
    This function enables you to disable the CA verification of the CKAN server.

    __Warning__:
    Only allow in a local environment!

    :return:
    """
    global allow_no_ca
    allow_no_ca = value

## Resource download from external urls
download_external_resource_urls:bool = True

def unlock_external_url_resource_download(value:bool=True) -> None:
    """
    This function enables the download of resources external from the CKAN server.

    :return:
    """
    global download_external_resource_urls
    download_external_resource_urls = value


## External code execution
# see: external_code_import

## Defining a data policy from an url
allow_policy_from_url:bool = False




#!python3
# -*- coding: utf-8 -*-
"""
Harvest from a mongo database using pymongo
"""
import argparse

from ckanapi_harvesters.harvesters.harvester_params import TableParams


class TableParamsMongoCollection(TableParams):
    """
    A table (CKAN DataStore) corresponds to a MongoDB collection.
    This subclass of TableParams implements an alias attribute for table name called collection.
    """
    def __init__(self, source: "TableParamsMongoCollection" =None):
        super().__init__(source)
        self.dbref_expand:bool = False
        if source is not None:
            source.copy(dest=self)

    # alias property for the table name setting: collection in MongoDB
    @property
    def collection(self) -> str:
        return self.table
    @collection.setter
    def collection(self, value: str):
        self.table = value

    def copy(self, *, dest=None):
        if dest is None:
            dest = TableParamsMongoCollection()
        super().copy(dest=dest)
        dest.dbref_expand = self.dbref_expand
        return dest

    @staticmethod
    def setup_cli_harvester_parser(parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        parser = TableParams.setup_cli_harvester_parser(parser=parser)
        parser.add_argument("--collection", type=str,
                            help="MongoDB collection name")  # normally specified in the File/URL attribute of builder
        # parser.add_argument("--table", help=argparse.SUPPRESS)  # do not display in help ==> conflict
        parser.add_argument("--dbref-expand",
                            help="Option to expand DBRefs",
                            action="store_true", default=False)  # applies to data cleaner
        return parser

    def initialize_from_cli_args(self, args: argparse.Namespace, base_dir: str = None, error_not_found: bool = True,
                                 default_proxies: dict = None, proxy_headers: dict = None) -> None:
        super().initialize_from_cli_args(args, base_dir=base_dir, error_not_found=error_not_found,
                                         default_proxies=default_proxies, proxy_headers=proxy_headers)
        self.dbref_expand = args.dbref_expand
        if args.collection is not None:
            self.collection = args.collection


if __name__ == '__main__':
    sample_instance = TableParamsMongoCollection()
    print("MongoDB harvester CLI-format options:")
    sample_instance.print_help_cli()



#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements the ckan connection definition.
"""
from typing import Union
from collections import OrderedDict

import pandas as pd

from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.path import make_path_relative
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element
from ckanapi_harvesters.auxiliary.ckan_auxiliary import ca_file_rel_to_dir, ca_arg_to_str
from ckanapi_harvesters.auxiliary.proxy_config import ProxyConfig
from ckanapi_harvesters.policies.data_format_policy import CkanPackageDataFormatPolicy


class BuilderCkan:
    def __init__(self, url:str=None, apikey_file:str=None, proxy:ProxyConfig=None):
        if proxy is None:
            proxy = ProxyConfig()
        self.url: str = url
        self.apikey_file: str = apikey_file
        self._proxy_config: ProxyConfig = proxy
        self._policy_file: Union[str, None] = None
        self._policy: Union[CkanPackageDataFormatPolicy, None] = None
        self.ckan_ca: Union[bool, str, None] = None
        self.extern_ca: Union[bool, str, None] = None
        self._ckan_ca_src: Union[str, None] = None
        self._extern_ca_src: Union[str, None] = None
        self.options_string: Union[str, None] = None
        self.limit: Union[int, None] = None
        self.time_between_requests: Union[float, None] = None
        self.default_thread_count: int = 1
        self.comment: Union[str, None] = None

    def __str__(self):
        return f"CKAN builder"

    def __copy__(self):
        return self.copy()

    def copy(self) -> "BuilderCkan":
        dest = BuilderCkan()
        dest.url = self.url
        dest.apikey_file = self.apikey_file
        dest._proxy_config = self._proxy_config.copy()
        dest._policy_file = self._policy_file
        if self._policy is not None:
            dest._policy = self._policy.copy()
        dest.ckan_ca = self.ckan_ca
        dest.extern_ca = self.extern_ca
        dest._ckan_ca_src = self._ckan_ca_src
        dest._extern_ca_src = self._extern_ca_src
        dest.options_string = self.options_string
        dest.limit = self.limit
        dest.time_between_requests = self.time_between_requests
        dest.default_thread_count = self.default_thread_count
        dest.comment = self.comment
        return dest

    @property
    def proxies(self) -> dict:
        return self._proxy_config.proxies
    @proxies.setter
    def proxies(self, proxies:dict) -> None:
        self._proxy_config.proxies = proxies
    @property
    def proxy_string(self) -> str:
        return self._proxy_config.proxy_string
    @proxy_string.setter
    def proxy_string(self, proxies:str) -> None:
        self._proxy_config.proxy_string = proxies

    @property
    def policy_file(self) -> str:
        return self._policy_file
    def set_policy_file(self, policy_file:str, *, ckan:CkanApi=None, base_dir:str=None, proxies:dict=None,
                        error_not_found:bool=True) -> None:
        if proxies is None:
            proxies = self._proxy_config.proxies
        self._policy_file = policy_file
        if policy_file is not None:
            self._policy = None
            if ckan is None:
                ckan = self.init_ckan(base_dir=base_dir)  # initiate a temporary ckan object to enable the load of the default policy
                # self._policy = CkanPackageDataFormatPolicy.from_json(policy_file, base_dir=base_dir, proxies=proxies, error_not_found=error_not_found)
            self._policy = ckan.load_policy(policy_file, base_dir=base_dir, proxies=proxies, error_not_found=error_not_found)
            self._policy_file = self._policy.source_file
        else:
            self._policy = None
    @property
    def policy(self) -> CkanPackageDataFormatPolicy:
        return self._policy

    def _load_from_df(self, ckan_df: pd.DataFrame, base_dir: str,
                      proxies:dict, error_not_found:bool=True) -> None:
        """
        Function to load builder parameters from a DataFrame, usually from an Excel worksheet

        :param ckan_df:
        :return:
        """
        ckan_df.columns = ckan_df.columns.map(str.lower)
        ckan_df.columns = ckan_df.columns.map(str.strip)
        # order is important here:
        if "ckan url" in ckan_df.columns:
            self.url = _string_from_element(ckan_df.pop("ckan url"))
        if "ckan api key file" in ckan_df.columns:
            self.apikey_file = _string_from_element(ckan_df.pop("ckan api key file"))
        if "proxies" in ckan_df.columns:
            self._proxy_config.proxy_string = _string_from_element(ckan_df.pop("proxies"))
        if "proxy authentication file" in ckan_df.columns:
            proxy_auth_file = _string_from_element(ckan_df.pop("proxy authentication file"))
            if proxy_auth_file is not None:
                self._proxy_config.load_proxy_auth_from_file(proxy_auth_file, base_dir=base_dir)
        self.ckan_ca = None
        self._ckan_ca_src = None
        if "ckan remote ca" in ckan_df.columns:
            ca_cert = _string_from_element(ckan_df.pop("ckan remote ca"))
            self.ckan_ca, self._ckan_ca_src = ca_file_rel_to_dir(ca_cert, base_dir)
        self.extern_ca = None
        self._extern_ca_src = None
        if "external ca" in ckan_df.columns:
            ca_cert = _string_from_element(ckan_df.pop("external ca"))
            self.extern_ca, self._extern_ca_src = ca_file_rel_to_dir(ca_cert, base_dir)
        if "data format policy file" in ckan_df.columns:
            policy_file = _string_from_element(ckan_df.pop("data format policy file"))
            self.set_policy_file(policy_file, base_dir=base_dir, proxies=proxies,
                                 error_not_found=error_not_found)
        if "options" in ckan_df.columns:
            self.options_string = _string_from_element(ckan_df.pop("options"))
        if "limit" in ckan_df.columns:
            number_str = _string_from_element(ckan_df.pop("limit"))
            if number_str is not None:
                self.limit = int(number_str)
        if "time between requests" in ckan_df.columns:
            number_str = _string_from_element(ckan_df.pop("time between requests"))
            if number_str is not None:
                self.time_between_requests = float(number_str)
        if "thread count" in ckan_df.columns:
            number_str = _string_from_element(ckan_df.pop("thread count"))
            if number_str is not None:
                self.default_thread_count = int(number_str)
        if "comment" in ckan_df.columns:
            self.comment = _string_from_element(ckan_df.pop("comment"))

    def _to_dict(self, base_dir:str) -> dict:
        """
        Function to export builder parameters to an Excel worksheet, using the same fields as the input format

        :see: _load_from_df
        :see: to_xls
        :return:
        """
        ckan_dict = OrderedDict()
        ckan_dict["CKAN URL"] = self.url
        ckan_dict["CKAN API key file"] = self.apikey_file
        ckan_dict["Proxies"] = self._proxy_config.proxy_string
        ckan_dict["Proxy authentication file"] = make_path_relative(self._proxy_config.proxy_auth_file, base_dir)
        ckan_dict["CKAN remote CA"] = ca_arg_to_str(self.ckan_ca, base_dir=base_dir, source_string=self._ckan_ca_src)
        ckan_dict["External remote CA"] = ca_arg_to_str(self.extern_ca, base_dir=base_dir, source_string=self._extern_ca_src)
        ckan_dict["Data format policy file"] = make_path_relative(self.policy_file, base_dir)
        ckan_dict["Options"] = self.options_string
        ckan_dict["Limit"] = self.limit
        ckan_dict["Time between requests"] = self.time_between_requests
        ckan_dict["Thread count"] = self.default_thread_count
        ckan_dict["Comment"] = self.comment
        return ckan_dict

    def _get_builder_df_help_dict(self) -> dict:
        ckan_help_dict = {
            "CKAN URL": "URL of the CKAN server e.g. https://demo.ckan.org/",
            "CKAN API key file": "Path to a file containing the API key in the first line",
            "Proxies": 'Proxies configuration, either one unique url or {"http": "http://proxy:8082", "https": "http://proxy:8082"}',
            "Proxy authentication file": "Path to a text file containing 3 lines with the proxy authentication method, username and password, relative to this Excel workbook folder. "
                                         + "This applies to all connexions (to CKAN server and external resources)",
            "CKAN remote CA": "Path to a custom CA certificate for the CKAN server (.pem), relative to this Excel workbook folder",
            "External CA": "Path to a custom CA certificate used for connexions other than the CKAN server, relative to this Excel workbook folder (.pem)",
            "Data format policy file": "Path to a JSON file containing the CKAN data format policy, relative to this Excel workbook folder",
            "Options": "List of options to initialize the CKAN API object in CLI format",
            "Limit": "Maximum number of records to return/send per request",
            "Time between requests": "Delay between upload/download requests (seconds) - recommended: 0.1 seconds",
            "Thread count": "Number of threads to use to upload/download large datasets",
        }
        return ckan_help_dict

    def _load_from_dict(self, ckan_dict: dict, base_dir: str, proxies:dict=None) -> None:
        ckan_df = pd.DataFrame([ckan_dict], index=["Value"])
        ckan_df = ckan_df.transpose()
        ckan_df.index.name = "Attribute"
        ckan_df = ckan_df.transpose()
        self._load_from_df(ckan_df, base_dir=base_dir, proxies=proxies)

    def _get_builder_df(self, base_dir:str) -> pd.DataFrame:
        """
        Converts the result of method _to_dict() into a DataFrame

        :return:
        """
        ckan_dict = self._to_dict(base_dir=base_dir)
        ckan_help_dict = self._get_builder_df_help_dict()
        ckan_df = pd.DataFrame([ckan_dict, ckan_help_dict], index=["Value", "Help"])
        ckan_df = ckan_df.transpose()
        ckan_df.index.name = "Attribute"
        return ckan_df

    def from_ckan(self, ckan: CkanApi) -> None:
        """
        Initialize fields from a CKAN instance.
        """
        self.url = ckan.url
        self.apikey_file = ckan.apikey.apikey_file
        self._proxy_config = ckan.params._proxy_config
        self.ckan_ca = ckan.params.ckan_ca
        self.extern_ca = ckan.params.extern_ca
        if ckan.policy is not None and ckan.policy_source is not None:
            self.set_policy_file(ckan.policy_source)

    def init_ckan(self, base_dir:str, ckan:CkanApi=None, default_proxies:dict=None,
                  proxies:Union[str,dict,ProxyConfig]=None) -> CkanApi:
        """
        Initialize a CKAN instance, following the parameters of the Excel workbook.
        The parameters from Excel have precedence on the values already contained in the CKAN object.
        However, the Excel workbook might not contain sufficient information.

        :param base_dir:
        :param ckan:
        :param default_proxies:
        :param proxies:
        :return:
        """
        if ckan is None:
            ckan = CkanApi(url=self.url)
        if self.url is not None:
            ckan.url = self.url.strip()
        if self.apikey_file is not None:
            ckan.load_apikey(self.apikey_file, base_dir=base_dir)
        if self.ckan_ca is not None:
            ckan.ckan_ca = self.ckan_ca
        if self.extern_ca is not None:
            ckan.extern_ca = self.extern_ca
        if proxies is not None:
            # proxies given by argument are prioritary over those specified in the builder
            ckan.set_proxies(proxies)
        elif self._proxy_config.is_defined():
            ckan._proxy_config = self._proxy_config
            ckan._proxy_config.replace_default_proxy(default_proxies)
        elif default_proxies is not None:
            ckan._proxy_config.proxies = default_proxies
        if self.policy is not None:
            ckan.policy = self.policy
        if self.options_string is not None:
            ckan.initialize_from_options_string(self.options_string,
                                                base_dir=base_dir, default_proxies=default_proxies)
        if self.limit is not None:
            ckan.set_limits(self.limit)
        if self.time_between_requests is not None:
            ckan.params.multi_requests_time_between_requests = self.time_between_requests
        return ckan




#!python3
# -*- coding: utf-8 -*-
"""
User custom IO function examples
"""
from typing import Union, Dict, List, Generator, Tuple
import io
from contextlib import contextmanager

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_model import CkanField, CkanResourceInfo
from ckanapi_harvesters.harvesters.file_formats.user_format import UserFileFormat


def read_function_example_df(file_path_or_stream:Union[str, io.IOBase], *, fields: Union[Dict[str, CkanField],None],
                             allow_chunks:bool=True, params:UserFileFormat = None, **kwargs) \
        -> Union[pd.DataFrame, List[dict]]:
    """
    Read a file/IO stream and return a unique DataFrame.
    This case is the simplest implementation.
    """
    return pd.read_csv(file_path_or_stream)

def read_function_example_df_with_metadata(file_path_or_stream:Union[str, io.IOBase], *, fields: Union[Dict[str, CkanField],None],
                                           allow_chunks:bool=True, params:UserFileFormat = None, **kwargs) \
        -> Tuple[Union[pd.DataFrame, List[dict]], CkanResourceInfo]:
    """
    Read a file/IO stream and return a unique DataFrame.
    This case returns as well metadata which can be read from the file.
    """
    return pd.read_csv(file_path_or_stream), CkanResourceInfo()

# use of a context manager when returning a custom DataFrame iterator in order to properly close the file if the process is interrupted (use of a with statement)
@contextmanager
def read_function_example_by_chunks(file_path_or_stream:Union[str, io.IOBase], *, fields: Union[Dict[str, CkanField],None],
                                    allow_chunks:bool=True, params:UserFileFormat = None, **kwargs) \
        -> Generator:
    """
    Read a file/IO stream and return a DataFrame generator.
    This function implements a context manager that ensures the file is closed properly when it is released.
    The DataFrame generator must be defined in a sub-function, such as in this example.

    Implementation prototype
    --------
    file_handle = open(file_path_or_stream, 'r')
    try:
        yield read_function_example_by_chunks_generator(file_handle)
    finally:
        file_handle.close()
    """
    file_handle = open(file_path_or_stream, 'r')
    try:
        yield read_function_example_by_chunks_generator(file_handle)
    finally:
        file_handle.close()

def read_function_example_by_chunks_generator(file_handle) -> Generator[Union[pd.DataFrame, List[dict]], None, None]:
    """
    This is the function which properly yields DataFrame chunks. It is called by read_function_example_by_chunks.
    """
    for df_chunk in pd.read_csv(file_handle, chunksize=100):
        yield df_chunk


def write_function_example(df: Union[pd.DataFrame, List[dict]], file_path_or_buffer:Union[str, io.IOBase],
                           *, fields: Union[Dict[str, CkanField],None], append:bool=False,
                           params:UserFileFormat = None, **kwargs) -> None:
    """
    This function writes a DataFrame to the given file path.
    """
    mode = 'a' if append else 'w'
    df.to_csv(file_path_or_buffer, mode=mode, index=False)



#!python3
# -*- coding: utf-8 -*-
"""
Setting the proxy from simple command line arguments
"""
import urllib.request
from typing import Union, Tuple
import os
import argparse
from warnings import warn
import copy
import json

from ckanapi_harvesters.auxiliary.ckan_defs import environ_keyword
from ckanapi_harvesters.auxiliary.path import sanitize_path, path_rel_to_dir
from ckanapi_harvesters.auxiliary.login import Login

import requests
from requests.auth import AuthBase, HTTPBasicAuth  #, HTTPProxyAuth

PROXY_AUTH_ENVIRON = "PROXY_AUTH_FILE"

class HttpsProxyDefError(Exception):
    def __init__(self):
        super().__init__("Only one of http_proxy or https_proxy is set")

def get_proxies_from_environ() -> dict:
    proxies = urllib.request.getproxies()
    return proxies
    # http_proxy = os.environ.get("http_proxy")
    # https_proxy = os.environ.get("https_proxy")
    # no_proxy = os.environ.get("no_proxy")
    # if http_proxy is not None and https_proxy is not None:
    #     proxies = {"http": http_proxy, "https": https_proxy}
    # elif http_proxy is not None:
    #     proxies = {"http": http_proxy, "https": http_proxy}
    # elif https_proxy is not None:
    #     raise HttpsProxyDefError()
    # else:
    #     proxies = None
    # if proxies is not None and no_proxy is not None:
    #     proxies["no"] = no_proxy
    # return proxies

def host_port_sep(url:Union[str,None], *, default_port:int=None) -> Tuple[Union[str,None],Union[int,None]]:
    if url is None:
        return None, None
    if ':' in url:
        host_prefix, host_suffix = "", ""
        if '@' in url:
            host_prefix, url = url.split('@')
        if '?' in url:
            url, host_suffix = url.split('?')
        host, port_str = url.split(':')
        host = host_prefix + host + host_suffix
        port = int(port_str)
    else:
        host, port = url, default_port
    return host, port

def _define_proxies(proxy_string:Union[str, dict], default_proxies:dict=None) -> dict:
    if proxy_string is None:
        proxies = None
    elif isinstance(proxy_string, dict):
        proxies = proxy_string
    elif isinstance(proxy_string, str):
        proxy_string = proxy_string.strip()
        proxy_mode = proxy_string.lower()
        if proxy_mode == environ_keyword:
            proxies = get_proxies_from_environ()
            if proxies is None:
                proxies = default_proxies
        elif proxy_mode == "unspecified":
            proxies = None  # do not specify the proxies - is equivalent to "environ"
        elif proxy_mode == "noproxy":
            proxies = {"http": "", "https": ""}  # do not use any proxy
        elif proxy_mode == "default":
            proxies = default_proxies  # default proxies, provided in argument
        elif proxy_string.startswith('{'):
            # proxy string is a string representation of proxy dictionary
            proxies = json.loads(proxy_string)
        else:
            # suppose string contains an url to a proxy server
            proxies = {"http": proxy_string, "https": proxy_string}
            # if "http" not in proxy_string:
            #     # url without http
            #     proxies = {"http": f"http://{proxy_string}", "https": f"http://{proxy_string}"}
            # else:
            #     proxies = {"http": proxy_string, "https": proxy_string}
    else:
        raise TypeError("proxy must be str or dict")
    return proxies


class ProxyConfig:
    def __init__(self, proxy_string:Union[str,dict]=None, default_proxies:dict=None,
                 proxy_headers:dict = None, proxy_auth:Union[AuthBase, Tuple[str,str]]=None) -> None:
        """
        :param proxy_string: string or proxies dict or ProxyConfig object.
        If a string is provided, it must be an url to a proxy or one of the following values:
            - "environ": use the proxies specified in the environment variables "http_proxy" and "https_proxy"
            - "noproxy": do not use any proxies
            - "unspecified": do not specify the proxies
            - "default": use value provided by default_proxies
        :param default_proxies: proxies used if proxies="default"
        :param proxy_headers: headers used to access the proxies, generally for authentication
        """
        if proxy_headers is None: proxy_headers = {}
        self._proxy_string:Union[str, dict, None] = None
        self._proxies:Union[dict,None] = None
        self._is_defined:bool = False
        self._default_proxies:Union[dict,None] = default_proxies
        self.proxy_headers: dict = proxy_headers
        self._proxy_auth: Union[AuthBase, Tuple[str,str], None] = proxy_auth
        self.proxy_auth_file: Union[str,None] = None
        self.proxy_auth_from_env: bool = False
        self.proxy_string = proxy_string  # property
        # self.load_proxy_auth_environ(error_not_found=False)  # recommended to base these parameters on user demand (confirm if there is a risk of leakage)

    def __str__(self):
        return str(self._proxies)

    def __copy__(self):
        return self.copy()

    def copy(self) -> "ProxyConfig":
        dest = ProxyConfig()
        dest._proxies = copy.deepcopy(self._proxies)
        dest._proxy_string = copy.deepcopy(self._proxy_string)
        dest._default_proxies = copy.deepcopy(self._default_proxies)
        dest.proxy_headers = copy.deepcopy(self.proxy_headers)
        dest._proxy_auth = copy.deepcopy(self._proxy_auth)
        dest._is_defined = self._is_defined
        dest.proxy_auth_file = self.proxy_auth_file
        dest.proxy_auth_from_env = self.proxy_auth_from_env
        return dest


    @property
    def proxy_string(self) -> Union[str, dict, None]:
        return self._proxy_string
    @proxy_string.setter
    def proxy_string(self, proxy_string:Union[str, dict, None]):
        self._proxy_string = proxy_string
        self._is_defined = proxy_string is not None
        self._proxies = _define_proxies(proxy_string, default_proxies=self._default_proxies)
    @property
    def proxies(self) -> dict:
        return self._proxies
    @proxies.setter
    def proxies(self, proxies:dict):
        self._proxy_string = proxies
        self._proxies = proxies
        self._is_defined = True
    @property
    def proxy_auth(self) -> Union[AuthBase, Tuple[str,str]]:
        return self._proxy_auth
    @proxy_auth.setter
    def proxy_auth(self, proxy_auth:Union[AuthBase, Tuple[str,str]]):
        self._proxy_auth = proxy_auth
        self.proxy_auth_file = None
        self.proxy_auth_from_env = False

    def get_host_port(self) -> Tuple[Union[str,None],Union[int,None]]:
        # special mode
        if self._proxies is None:
            return None, None
        elif "http_proxy" in self._proxies.keys() and self._proxies["http_proxy"] is not None:
            return host_port_sep(self._proxies["http_proxy"])
        else:
            return None, None

    def get_proxy_login(self) -> Login:
        if self._proxy_auth is None:
            return Login()
        else:
            assert(isinstance(self._proxy_auth, HTTPBasicAuth))  # HTTPProxyAuth is a super class of HTTPBasicAuth
            return Login(self._proxy_auth.username, self._proxy_auth.password)

    @staticmethod
    def from_str_or_config(proxies:Union[str,dict, "ProxyConfig"],
                           *, default_proxies:dict=None, proxy_headers:dict=None) -> "ProxyConfig":
        if proxies is None:
            return ProxyConfig(None, default_proxies=default_proxies , proxy_headers=proxy_headers)
        elif isinstance(proxies, ProxyConfig):
            if proxy_headers is not None:
                proxies.proxy_headers = proxy_headers
            return proxies
        else:
            return ProxyConfig(proxies, default_proxies=default_proxies, proxy_headers=proxy_headers)

    def replace_default_proxy(self, default_proxies:dict) -> None:
        if self._proxy_string is not None and self._proxy_string.lower() == "default":
            self._proxies = default_proxies

    def reset(self) -> None:
        self._proxy_string = None
        self._proxies = None
        self._is_defined = False

    def is_defined(self) -> bool:
        return self._is_defined

    def load_proxy_auth_environ(self, *, error_not_found:bool=False) -> bool:
        proxy_auth_file = sanitize_path(os.environ.get(PROXY_AUTH_ENVIRON))  # "PROXY_AUTH_FILE"
        if proxy_auth_file is not None:
            proxy_keyword = proxy_auth_file.strip().lower()
            assert(not proxy_keyword == environ_keyword)  # this value would create an infinite loop
            if self.load_proxy_auth_from_file(proxy_auth_file, error_not_found=error_not_found):
                self.proxy_auth_from_env = True
                return True
        return False

    def load_proxy_auth_from_file(self, file_path:str, *, base_dir:str=None, error_not_found:bool=True) -> bool:
        file_path = path_rel_to_dir(file_path, base_dir=base_dir, keyword_exceptions={environ_keyword})
        proxy_keyword = file_path.strip().lower()
        if proxy_keyword == environ_keyword:
            # this keyword is not very useful if proxy authentication file is loaded from environment anyway
            return self.load_proxy_auth_environ(error_not_found=error_not_found)
        if (not error_not_found) and (not os.path.isfile(file_path)):
            msg = f"Proxy authentication file does not exist: {file_path}"
            warn(msg)
            return False
        self.proxy_auth_file = file_path
        self.proxy_auth_from_env = False
        with open(file_path, "r") as f:
            auth_type = f.readline().strip().lower()
            username = f.readline().strip()
            password = f.readline().strip()
            if auth_type == "basic" or auth_type == "httpbasicauth":
                self._proxy_auth = requests.auth.HTTPBasicAuth(username, password)
            elif auth_type == "proxy" or auth_type == "httpproxyauth":
                self._proxy_auth = requests.auth.HTTPProxyAuth(username, password)
            elif auth_type == "digest" or auth_type == "httpdigestauth":
                self._proxy_auth = requests.auth.HTTPDigestAuth(username, password)
            elif auth_type == "none":
                self._proxy_auth = None
            else:
                raise KeyError(f"Unknown auth type {auth_type}")
        return True


    @staticmethod
    def _setup_cli_proxy_parser(parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        """
        Define or add CLI arguments to initialize the proxy
        parser help message:

        Proxy parameters initialization

        options:
          -h, --help            show this help message and exit
          --proxy PROXY         Proxy for HTTP and HTTPS

        :param parser: option to provide an existing parser to add the specific fields needed to initialize a CKAN API connection
        :return:
        """
        if parser is None:
            parser = argparse.ArgumentParser(description="Proxy parameters initialization", add_help=False)
        parser.add_argument("--proxy", type=str,
                            help="Proxy for HTTP and HTTPS")
        parser.add_argument("--http-proxy", type=str,
                            help="HTTP proxy")
        parser.add_argument("--https-proxy", type=str,
                            help="HTTPS proxy")
        parser.add_argument("--no-proxy", type=str,
                            help="Proxy exceptions")
        parser.add_argument("--proxy-auth-file", type=str,
                            help="Path to a proxy authentication file with 3 lines (authentication method, username, password)")
        return parser

    @staticmethod
    def from_cli_args(args: argparse.Namespace, *, base_dir:str=None, error_not_found:bool=True,
                      default_proxies:dict=None, proxy_headers:dict=None) -> "ProxyConfig":
        proxy_string, proxies = None, None
        if args.proxy is not None:
            proxy_string = args.proxy
        elif args.http_proxy is not None and args.https_proxy is not None:
            proxies = {"http": args.http_proxy, "https": args.https_proxy}
        elif args.http_proxy is not None:
            proxies = {"http": args.http_proxy, "https": args.http_proxy}
        if proxies is not None and args.no_proxy is not None:
            proxies["no"] = args.no_proxy
        elif args.https_proxy is not None:
            raise HttpsProxyDefError()
        if proxy_string is not None:
            proxy_config = ProxyConfig(proxy_string, default_proxies=default_proxies, proxy_headers=proxy_headers)
        elif proxies is not None:
            proxy_config = ProxyConfig(proxies, default_proxies=default_proxies, proxy_headers=proxy_headers)
        else:
            proxy_config = None
        if args.proxy_auth_file is not None:
            if proxy_config is not None:
                proxy_config.load_proxy_auth_from_file(args.proxy_auth_file, base_dir=base_dir, error_not_found=error_not_found)
            else:
                raise Exception(f"Proxy authentication file specified without proxy specification: {args.proxy_auth_file}")
        return proxy_config




#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements the basic resources. See builder_datastore for specific functions to initiate datastores.
"""
from typing import Union, Any
from abc import ABC, abstractmethod
import os
from warnings import warn
import copy
import io
from collections import OrderedDict
import argparse

import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.auxiliary.ckan_auxiliary import upload_prepare_requests_files_arg, assert_or_raise
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo
from ckanapi_harvesters.auxiliary.ckan_progress_callbacks import CkanProgressCallback
from ckanapi_harvesters.auxiliary.path import resolve_rel_path
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, _bool_from_string
from ckanapi_harvesters.auxiliary.ckan_model import CkanState
from ckanapi_harvesters.auxiliary.ckan_errors import CkanArgumentError, MissingIdError, FunctionMissingArgumentError, MandatoryAttributeError
from ckanapi_harvesters.builder.builder_errors import ResourceFileNotExistMessage, EmptyPackageNameException


builder_request_default_auth_if_ckan:Union[bool,None] = True  # fill authentification headers for requests with CkanApi requests proxy method if same domain is used by default
builder_default_resource_state:CkanState = CkanState.Active  # CkanState.Draft
initial_resource_building_state:Union[CkanState,None] = None  # CkanState.Draft


class BuilderResourceABC(ABC):
    def __init__(self, *, parent:"BuilderPackage", name:str=None, format:str=None, description:str=None,
                 state:CkanState=None, enable_download:bool=True,
                 resource_id:str=None, download_url:str=None, options_string:str=None):
        if options_string is None:
            options_string = ""
        if state is None:
            state = builder_default_resource_state
        self.name: Union[str,None] = name
        self.resource_attributes: Union[CkanResourceInfo,None] = None
        self.resource_attributes_user: CkanResourceInfo = CkanResourceInfo()
        self.resource_attributes_data_source: Union[CkanResourceInfo,None] = None
        self.columns_sheet_name: Union[str, None] = None
        self.resource_attributes_user.format = format
        self.resource_attributes_user.description = description
        self.resource_attributes_user.state = state
        self.enable_download:bool = enable_download
        self.options_string: str = options_string
        self.comment: Union[str,None] = None
        # Map information, if present
        self.parent_package: "BuilderPackage" = parent  # BuilderPackage
        self.download_url: Union[str,None] = download_url
        self.known_id: Union[str,None] = resource_id  # known id is simply indicative. ckan.map is the reference. User can override CKAN API by calling BuilderPackage.update_ckan_map at his own risks.
        self.known_resource_info: Union[CkanResourceInfo,None] = None
        # Functions inputs/outputs
        self.sample_data_source: str = ""
        self.reupload_on_update: bool = True
        self.downloaded_destination: str = ""
        self.download_skip_existing:bool = True  # True: do not overwrite files
        self.download_error_not_found:bool = True
        self.create_default_view:bool = True
        self.progress_callback = CkanProgressCallback()

    def __copy__(self):
        return self.copy()

    @property
    def package_name(self):
        """
        Returns the package name of the parent package.
        You cannot assign the package name through this property.
        Setting this property only performs a check. This will be removed in future releases.
        To change package name, change the package_name attribute of the parent_package.
        """
        return self.parent_package.package_name
    @package_name.setter
    def package_name(self, value):
        # remove in future releases
        assert_or_raise(value == self.parent_package.package_name, RuntimeError())

    def clear_secrets_and_disconnect(self) -> None:
        return

    def __del__(self):
        self.clear_secrets_and_disconnect()

    # Context Manager behavior ----------
    # to use ResourceBuilder in a "with" statement
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.clear_secrets_and_disconnect()
        return True
    # -------------

    @abstractmethod
    def copy(self, *, dest:"BuilderResourceABC"=None, parent:"BuilderPackage"=None):
        if parent is None:
            parent = self.parent_package
        dest.name = self.name
        dest.parent_package = parent  # may replace parent assigned in constructor from current instance
        # dest.package_name = self.package_name  # double-check is not necessary, especially if parent changed
        dest.columns_sheet_name = self.columns_sheet_name
        dest.resource_attributes = self.resource_attributes.copy() if self.resource_attributes is not None else None
        dest.resource_attributes_user = self.resource_attributes_user.copy()
        dest.resource_attributes_data_source = self.resource_attributes_data_source.copy() if self.resource_attributes_data_source is not None else None
        dest.enable_download = self.enable_download
        dest.options_string = self.options_string
        dest.known_id = self.known_id
        dest.download_url = self.download_url
        dest.comment = self.comment
        dest.sample_data_source = self.sample_data_source
        dest.reupload_on_update = self.reupload_on_update
        dest.downloaded_destination = self.downloaded_destination
        dest.download_skip_existing = self.download_skip_existing
        dest.download_error_not_found = self.download_error_not_found
        dest.create_default_view = self.create_default_view
        return dest

    def _check_mandatory_attributes(self):
        if self.name is None:
            raise MandatoryAttributeError("Resource", "name")

    def init_options_from_ckan(self, ckan:CkanApi, *, base_dir:str=None) -> None:
        """
        Function to initialize some parameters from the ckan object
        """
        if self.known_resource_info is None:
            self.known_resource_info = ckan.get_resource_info_or_request(self.name, self.package_name, error_not_found=False)
        self._update_metadata(ckan, base_dir=base_dir)

    def _apply_options(self, base_dir: str = None) -> None:
        return

    def _update_metadata(self, ckan: CkanApi, *, base_dir:str=None) -> None:
        """
        Function to initialize metadata from the data source.
        The attribute self.known_resource_info must be queried before this call
        Examples for a DataStore:
            - List of fields
            - Detect field types from example DataFrame
            - Add descriptions from data source
        """
        self.initialize_from_options_string(base_dir=base_dir)

    def initialize_from_options_string(self, base_dir:str, *, options_string:str=None, parser:argparse.ArgumentParser=None) -> None:
        pass

    @abstractmethod
    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        # abstract method because does not take into account file/url field
        self.name = _string_from_element(row["name"], empty_value="", strip=True)
        if self.name is None:
            raise MandatoryAttributeError("Resource", "name")
        self.columns_sheet_name = None
        if "datastore columns sheet" in row.keys():
            self.columns_sheet_name = _string_from_element(row["datastore columns sheet"])
        self.resource_attributes_user.format = _string_from_element(row["format"]).upper().strip()
        self.resource_attributes_user.description = None
        if "description" in row.keys():
            self.resource_attributes_user.description = _string_from_element(row["description"])
        self.enable_download = True
        if "options" in row.keys():
            self.options_string = _string_from_element(row["options"], empty_value="")
        if "download" in row.keys():
            self.enable_download = _bool_from_string(row["download"])
        self.resource_attributes_user.state = None
        if "state" in row.keys():
            state = _string_from_element(row["state"])
            if state is not None:
                self.resource_attributes_user.state = CkanState.from_str(state)
        # Map information, if present
        self.known_id = None
        self.download_url = None
        if "known id" in row.keys():
            self.known_id = _string_from_element(row["known id"])
        if "known url" in row.keys():
            self.download_url = _string_from_element(row["known url"])
        if "comment" in row.keys():
            self.comment = _string_from_element(row["comment"])

    def get_or_query_resource_id(self, ckan: CkanApi, cancel_if_present:bool=True, error_not_found:bool=True) -> str:
        """
        Store/retrieve resource ID in the class attributes.
        """
        package_name = self.package_name
        if package_name == "":
            raise EmptyPackageNameException()
        resource_id = ckan.get_resource_id_or_request(self.name, package_name=package_name, error_not_mapped=error_not_found)
        self.known_id = resource_id  # update
        return self.known_id

    def get_or_query_package_id(self, ckan: CkanApi) -> str:
        """
        Obtain package ID from the package name. This can lead to a request to the API.
        """
        package_name = self.package_name
        if package_name == "":
            raise EmptyPackageNameException()
        ckan.map_resources(package_name, only_missing=True)
        package_id = ckan.map.get_package_id(package_name=package_name)
        return package_id

    @staticmethod
    @abstractmethod
    def resource_mode_str() -> str:
        raise NotImplementedError()

    def __str__(self):
        return f"Resource builder for {self.name} in mode {self.resource_mode_str()}"

    @abstractmethod
    def _to_dict(self, include_id:bool=True) -> dict:
        d = OrderedDict([
            ("Name", self.name),
            ("DataStore Columns Sheet", self.columns_sheet_name),
            ("Format", self.resource_attributes_user.format if self.resource_attributes_user.format else ""),
            ("State", self.resource_attributes_user.state.name if self.resource_attributes_user.state is not None else ""),
            ("Mode", self.resource_mode_str()),
            ("File/URL", None),  # concrete implementations must fill this field
            ("Options", self.options_string),
            ("Download", str(self.enable_download)),
            ("Description", self.resource_attributes_user.description if self.resource_attributes_user.description else ""),
            ("Primary key", ""),
            ("Indexes", ""),
            ("Upload function", ""),
            ("Download function", ""),
            ("Aliases", ""),
            ("Comment", self.comment if self.comment else ""),
        ])
        if include_id and self.known_id is not None:
            d["Known ID"] = self.known_id
        if include_id and self.download_url is not None:
            d["Known URL"] = self.download_url
        return d

    def _merge_resource_attributes(self, *, override_ckan:bool=False):
        """
        Merge resource attributes into self.resource_attributes in the following priority order:
        1. Existing metadata from CKAN server. This can be ignored using the override_ckan argument.
        2. Metadata provided by the user in the Excel worksheet
        3. Metadata found automatically from the data source (e.g. in file header or database)
        """
        self.resource_attributes = self.resource_attributes_user.copy()
        if not override_ckan and self.known_resource_info is not None:
            self.resource_attributes.update_missing(self.known_resource_info)
        if self.resource_attributes_data_source is not None:
            self.resource_attributes.update_missing(self.resource_attributes_data_source)

    def _to_row(self) -> pd.Series:
        row = pd.Series(self._to_dict())
        row.index = row.index.map(str.lower)
        row.index = row.index.map(str.strip)
        return row

    @staticmethod
    @abstractmethod
    def sample_file_path_is_url() -> bool:
        raise NotImplementedError()

    @abstractmethod
    def get_sample_file_path(self, resources_base_dir:str, ckan:Union[CkanApi,None]) -> Union[str,None]:
        """
        Function returning the local resource file name for the sample file.

        :param resources_base_dir: base directory to find the resources on the local machine
        :return:
        """
        raise NotImplementedError()

    @abstractmethod
    def load_sample_data(self, resources_base_dir:str) -> Union[bytes,None]:
        """
        Function returning the data from the indicated resources.

        :param resources_base_dir: base directory to find the resources on the local machine
        :return:
        """
        raise NotImplementedError()

    @abstractmethod
    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        """
        Test the presence of the files/urls used in the upload/patch requests.

        :param resources_base_dir:
        :return: None if success, error message otherwise
        """
        raise NotImplementedError()

    @abstractmethod
    def patch_request(self, ckan:CkanApi, package_id:str, *,
                      reupload:bool=None, override_ckan:bool=False,
                      resources_base_dir:str=None) -> CkanResourceInfo:
        """
        Function to perform all the necessary requests to initiate/reupload the resource on the CKAN server.

        :param resources_base_dir:
        :param ckan:
        :param reupload: option to reupload the resource
        :return:
        """
        # TODO: call to API resource_patch
        # ckan.resource_patch
        raise NotImplementedError()

    def upload_request(self, resources_base_dir:str, ckan:CkanApi, package_id:str):
        # might be dead code
        # this function (patch_request) gets specialized in certain cases
        return self.patch_request(ckan, package_id, resources_base_dir=resources_base_dir, reupload=True)

    def upload_request_final(self, ckan:CkanApi, *, force:bool=False) -> None:
        if initial_resource_building_state is not None:
            resource_id = self.get_or_query_resource_id(ckan)
            ckan.resource_patch(resource_id, state=self.resource_attributes.state)

    @abstractmethod
    def download_resource_bytes(self, ckan:CkanApi, full_download:bool=True, **kwargs) -> bytes:
        """
        Download the resource and return the data as bytes.

        :param ckan:
        :param out_dir:
        :param full_download: Some resources like URLs are not downloaded by default. Large datasets are also limited to one request for this function by default.
        :param threads:
        :return:
        """
        raise NotImplementedError()

    def download_sample_df(self, ckan: CkanApi, *, limit: int = 100,
                           search_all: bool = False, download_alter: bool = False,
                           **kwargs) -> Union[pd.DataFrame, None]:
        return None

    @abstractmethod
    def download_request(self, ckan:CkanApi, out_dir:str, *, full_download:bool=True, force:bool=False, threads:int=1, return_data:bool=False) -> Any:
        """
        Download the resource and save in a file pointed by out_dir.
        In most implementations, this calls the download_resource_bytes method.

        :param ckan:
        :param out_dir:
        :param full_download: Some resources like URLs are not downloaded by default. Large datasets are treated with a multi-threaded approach.
        :param threads:
        :param force: option to bypass the enable_download attribute of resources
        :return:
        """
        raise NotImplementedError()

    def _to_ckan_resource_info(self, package_id:str, check_id:bool=True) -> CkanResourceInfo:
        """
        Return resource info object from the information of the Excel workbook.
        No requests are made but to use this data in the ckan object, the ID and name of the resource are mandatory.

        :param package_id:
        :param check_id:
        :return:
        """
        if self.known_id is None and check_id:
            msg = MissingIdError("resource", self.name)
            raise msg
        self._merge_resource_attributes(override_ckan=True)
        resource_info = CkanResourceInfo()
        resource_info.id = self.known_id
        resource_info.package_id = package_id
        resource_info.name = self.name
        resource_info.description = self.resource_attributes.description
        resource_info.state = self.resource_attributes.state
        resource_info.format = self.resource_attributes.format
        resource_info.download_url = self.download_url
        return resource_info

    def resource_info_request(self, ckan:CkanApi, error_not_found:bool=True) -> Union[CkanResourceInfo, None]:
        resource_id = self.get_or_query_resource_id(ckan, cancel_if_present=False, error_not_found=error_not_found)
        if resource_id is None and not self.download_error_not_found:
            return None
        res_info = ckan.get_resource_info_or_request(resource_id)
        self.known_id = resource_id
        return res_info

    def delete_request(self, ckan:CkanApi, *, error_not_found:bool=False):
        """
        Delete the resource from the CKAN server.

        :return:
        """
        resource_id = self.get_or_query_resource_id(ckan, error_not_found=error_not_found)
        if resource_id is not None:
            ckan.resource_delete(resource_id)


class BuilderFileABC(BuilderResourceABC, ABC):
    """
    Abstract class defining the behavior for a resource represented by a file (not a DataStore)
    """
    def __init__(self, *, parent:"BuilderPackage", name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, file_name:str=None):
        super().__init__(parent=parent, name=name, format=format, description=description, resource_id=resource_id, download_url=download_url)
        self.file_name: str = file_name

    def copy(self, *, dest=None, parent=None):
        super().copy(dest=dest, parent=parent)
        dest.file_name = self.file_name
        return dest

    def _check_mandatory_attributes(self):
        super()._check_mandatory_attributes()
        if self.file_name is None:
            raise MandatoryAttributeError(self.resource_mode_str(), "File")

    @abstractmethod
    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row, base_dir=base_dir)
        self.file_name = _string_from_element(row["file/url"], strip=True)

    def patch_request(self, ckan: CkanApi, package_id: str, *, reupload: bool = None, override_ckan:bool=False,
                      resources_base_dir:str=None,
                      payload:Union[bytes, io.BufferedIOBase]=None) -> CkanResourceInfo:
        """
        Perform a patch of the resource on the CKAN server.
        A patch is a full update of the metadata of the resource, and of the DataStore if appropriate.
        The source file of the resource is also uploaded (or a first file for large DataStores).

        :param ckan:
        :param package_id:
        :param reupload:
        :param resources_base_dir:
        :param payload:
        :return:
        """
        self._merge_resource_attributes(override_ckan=override_ckan)
        if reupload is None: reupload = self.reupload_on_update
        if payload is None:
            payload = self.load_sample_data(resources_base_dir=resources_base_dir)
        payload_file_name = self.file_name
        files = upload_prepare_requests_files_arg(payload=payload, payload_name=payload_file_name)
        res_info = ckan.resource_create(package_id, name=self.name, format=self.resource_attributes.format,
                                        description=self.resource_attributes.description,
                                        state=initial_resource_building_state if initial_resource_building_state is not None else self.resource_attributes.state,
                                        files=files, datastore_create=False, auto_submit=False, create_default_view=self.create_default_view,
                                        cancel_if_exists=True, update_if_exists=True, reupload=reupload,
                                        progress_callback=self.progress_callback)
        self.known_id = res_info.id
        self.upload_request_final(ckan)
        return res_info

    def download_resource_bytes(self, ckan:CkanApi, full_download:bool=True, search_all:bool=True, **kwargs) -> Union[bytes, None]:
        resource_id = self.get_or_query_resource_id(ckan=ckan, error_not_found=self.download_error_not_found)
        if resource_id is None and not self.download_error_not_found:
            return None
        resource_info, response = ckan.resource_download(resource_id)
        if response is not None:
            return response.content
        else:
            return None

    def download_request(self, ckan: CkanApi, out_dir: str, *, full_download:bool=True, threads:int=1,
                         force:bool=False, return_data:bool=False, **kwargs) -> None:
        if (not self.enable_download) and (not force):
            msg = f"Did not download resource {self.name} because download was disabled."
            warn(msg)
            return
        if out_dir is not None:
            self.downloaded_destination = resolve_rel_path(out_dir, self.file_name, field=f"File/URL of resource {self.name}")
            if self.download_skip_existing and os.path.exists(self.downloaded_destination):
                return
        content = self.download_resource_bytes(ckan=ckan, full_download=full_download, **kwargs)
        if out_dir is not None and content is not None:
            os.makedirs(out_dir, exist_ok=True)
            with open(self.downloaded_destination, "wb") as f:
                f.write(content)
                f.close()


# class BuilderResourceUnmanagedABC(BuilderResourceABC, ABC):
#     # dead code
#     def __init__(self, *, name:str=None, format:str=None, description:str=None,
#                  resource_id:str=None, download_url:str=None):
#         super().__init__(name=name, format=format, description=description, resource_id=resource_id, download_url=download_url)
#         self.file_name: str = name
#
#     def copy(self, *, dest=None):
#         super().copy(dest=dest)
#         dest.file_name = self.file_name
#         return dest
#
#     def _load_from_df_row(self, row: pd.Series):
#         super()._load_from_df_row(row=row, base_dir=base_dir)
#         self.file_name = self.name
#         self._check_mandatory_attributes()
#
#     def _to_dict(self, include_id:bool=True) -> dict:
#         d = super()._to_dict(include_id=include_id)
#         d["File/URL"] = ""
#         return d
#
#     def load_sample_data(self, resources_base_dir:str) -> bytes:
#         return None


class BuilderResourceUnmanaged(BuilderFileABC):  #, BuilderResourceUnmanagedABC):  # multiple inheritance can give undefined results
    """
    Class to manage a resource metadata without specifying its contents during the upload process.
    """
    def __init__(self, *, parent:"BuilderPackage", name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None):
        super().__init__(parent=parent, name=name, format=format, description=description, resource_id=resource_id, download_url=download_url)
        self.file_name: str = name
        self.default_payload: Union[bytes, io.BufferedIOBase, None] = None

    def copy(self, *, dest=None, parent=None):
        if dest is None:
            dest = BuilderResourceUnmanaged(parent=self.parent_package)
        super().copy(dest=dest, parent=parent)
        dest.file_name = self.file_name
        dest.default_payload = copy.deepcopy(self.default_payload)
        return dest

    @staticmethod
    def resource_mode_str() -> str:
        return "Unmanaged"

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row, base_dir=base_dir)
        self.file_name = self.name
        self._check_mandatory_attributes()

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = ""
        return d

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return False

    def get_sample_file_path(self, resources_base_dir:str, ckan:Union[CkanApi,None]=None) -> Union[str,None]:
        return None

    def load_sample_data(self, resources_base_dir:str) -> Union[bytes,None]:
        return None

    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[ContextErrorLevelMessage,None]:
        return None

    def patch_request(self, ckan:CkanApi, package_id:str, *,
                      reupload:bool=None, override_ckan:bool=False, resources_base_dir:str=None,
                      payload:Union[bytes, io.BufferedIOBase]=None) -> CkanResourceInfo:
        self._merge_resource_attributes(override_ckan=override_ckan)
        if payload is None:
            payload = self.default_payload
        if reupload is None: reupload = self.reupload_on_update and payload is not None
        payload_file_name = self.file_name
        files = upload_prepare_requests_files_arg(payload=payload, payload_name=payload_file_name) if payload is not None else None
        res_info = ckan.resource_create(package_id, name=self.name, format=self.resource_attributes.format,
                                        description=self.resource_attributes.description,
                                        state=initial_resource_building_state if initial_resource_building_state is not None else self.resource_attributes.state,
                                        files=files, datastore_create=False, auto_submit=False, create_default_view=self.create_default_view,
                                        cancel_if_exists=True, update_if_exists=True, reupload=reupload,
                                        progress_callback=self.progress_callback)
        self.known_id = res_info.id
        self.upload_request_final(ckan)
        return res_info


class BuilderFileBinary(BuilderFileABC):
    """
    Concrete implementation for a binary file.
    """
    def copy(self, *, dest=None, parent=None):
        if dest is None:
            dest = BuilderFileBinary(parent=self.parent_package)
        super().copy(dest=dest, parent=parent)
        return dest

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return False

    def get_sample_file_path(self, resources_base_dir:str, ckan:Union[CkanApi,None]=None) -> str:
        return resolve_rel_path(resources_base_dir, self.file_name, field=f"File/URL of resource {self.name}")

    def load_sample_data(self, resources_base_dir:str) -> bytes:
        self.sample_source = self.get_sample_file_path(resources_base_dir)
        with open(self.sample_source, "rb") as f:
            contents = f.read()
            f.close()
        return contents

    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        file_path = self.get_sample_file_path(resources_base_dir=resources_base_dir)
        if os.path.isfile(file_path):
            return None
        else:
            return ResourceFileNotExistMessage(self.name, ErrorLevel.Error, f"Missing file for resource {self.name}: {file_path}")

    @staticmethod
    def resource_mode_str() -> str:
        return "File"

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row, base_dir=base_dir)
        self._check_mandatory_attributes()

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = self.file_name
        return d


class BuilderUrlABC(BuilderFileABC, ABC):
    """
    Abstract behavior for a resource defined by an external URL.
    """
    def __init__(self, *, parent:"BuilderPackage", name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, url:str=None):
        super().__init__(parent=parent, name=name, format=format, description=description, resource_id=resource_id, download_url=download_url)
        self.url = url
        self.file_name: str = name

    def copy(self, *, dest=None, parent=None):
        super().copy(dest=dest, parent=parent)
        dest.url = self.url
        dest.file_name = self.file_name
        return dest

    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        if ckan is None:
            return ResourceFileNotExistMessage(self.name, ErrorLevel.Warning, "Could not determine if resource url exists because ckan argument was not provided.")
        else:
            return ckan.download_url_proxy_test_head(self.url, **kwargs)

    def _check_mandatory_attributes(self):
        super()._check_mandatory_attributes()
        if self.url is None:
            raise MandatoryAttributeError(self.resource_mode_str(), "URL")

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row, base_dir=base_dir)
        self.url: str = _string_from_element(row["file/url"], strip=True)
        self.file_name = self.name
        self._check_mandatory_attributes()

    def download_request(self, ckan: CkanApi, out_dir: str, *, full_download:bool=False, threads:int=1,
                         force:bool=False, return_data:bool=False, **kwargs) -> None:
        # do not download URLs by default
        if full_download:
            super().download_request(ckan=ckan, out_dir=out_dir,full_download=full_download, force=force,
                                     threads=threads, **kwargs)

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = self.url
        return d


class BuilderUrl(BuilderUrlABC):
    """
    Class for a resource defined by an external URL.
    """
    @staticmethod
    def resource_mode_str() -> str:
        return "URL"

    def copy(self, *, dest=None, parent=None):
        if dest is None:
            dest = BuilderUrl(parent=self.parent_package)
        super().copy(dest=dest, parent=parent)
        return dest

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return True

    def get_sample_file_path(self, resources_base_dir: str, ckan:Union[CkanApi,None]=None) -> str:
        return self.url

    def load_sample_data(self, resources_base_dir:str, *, ckan:CkanApi=None,
                         proxies:dict=None, headers:dict=None) -> bytes:
        self.sample_source = self.url
        if ckan is None:
            raise FunctionMissingArgumentError("BuilderDataStoreUrl.load_sample_data", "ckan")
        return ckan.download_url_proxy(self.url, proxies=proxies, headers=headers, auth_if_ckan=builder_request_default_auth_if_ckan).content

    def patch_request(self, ckan: CkanApi, package_id: str, *, reupload: bool = None, override_ckan:bool=False,
                      resources_base_dir:str=None,
                      payload:Union[bytes, io.BufferedIOBase]=None) -> CkanResourceInfo:
        self._merge_resource_attributes(override_ckan=override_ckan)
        if reupload is None: reupload = self.reupload_on_update
        if payload is not None:
            raise CkanArgumentError("payload", "resource defined from URL patch")
        resource_info = ckan.resource_create(package_id, name=self.name, format=self.resource_attributes.format,
                                    description=self.resource_attributes.description,
                                    state=initial_resource_building_state if initial_resource_building_state is not None else self.resource_attributes.state,
                                    url=self.url, auto_submit=False, datastore_create=False, create_default_view=self.create_default_view,
                                    cancel_if_exists=True, update_if_exists=True, reupload=reupload,
                                    progress_callback=self.progress_callback)
        self.upload_request_final(ckan)
        return resource_info


from ckanapi_harvesters.builder.builder_package import BuilderPackage




#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements functions to initiate a DataStore.
"""
from typing import List, Union, Generator
from collections import OrderedDict

import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.builder.builder_field import BuilderField
from ckanapi_harvesters.builder.builder_errors import ResourceFileNotExistMessage
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.auxiliary.ckan_model import UpsertChoice
from ckanapi_harvesters.builder.builder_resource_datastore_multi_folder import BuilderDataStoreFolder
from ckanapi_harvesters.builder.builder_resource_multi_abc import FileChunkDataFrame


class BuilderDataStoreCkan(BuilderDataStoreFolder):
    """
    Merge of existing CKAN DataStores (on the same server) into a single DataStore
    """
    def __init__(self, *, parent, name: str = None, format: str = None, description: str = None,
                 resource_id: str = None, download_url: str = None, file_name: str = None, options_string:str=None, base_dir:str=None):
        super().__init__(parent=parent, name=name, format=format, description=description, resource_id=resource_id,
                         download_url=download_url, dir_name="", options_string=options_string, base_dir=base_dir)
        self.upsert_method: UpsertChoice = UpsertChoice.Upsert
        self.resource_ids: List[str] = [resource_id.strip() for resource_id in ckan_tags_sep.split(file_name)]
        raise NotImplementedError()  # not tested

    def copy(self, *, dest=None, parent=None):
        if dest is None:
            dest = BuilderDataStoreCkan(parent=self.parent_package)
        super().copy(dest=dest, parent=parent)
        dest.resource_ids = self.resource_ids
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row, base_dir=base_dir)
        file_name_field: str = _string_from_element(row["file/url"], strip=True)
        if file_name_field is not None:
            self.resource_ids = [resource_id.strip() for resource_id in ckan_tags_sep.split(file_name_field)]
        else:
            self.resource_ids = []

    @staticmethod
    def resource_mode_str() -> str:
        return "CKAN DataStore merge"

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return True

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = ckan_tags_sep.join(self.resource_ids)
        return d

    ## upload chunks ----------------
    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        # check resource ids exist on CKAN
        missing_resources = []
        for resource_id in self.resource_ids:
            if ckan.get_resource_info_or_request(resource_id, error_not_found=False) is None:
                missing_resources.append(resource_id)
        if len(missing_resources) == 0:
            return None
        else:
            return ResourceFileNotExistMessage(self.name, ErrorLevel.Error, f"Missing source resources for {self.name}: {','.join(missing_resources)}")

    def get_sample_file_path(self, resources_base_dir:str, ckan:Union[CkanApi,None]=None, file_index:int=0) -> str:
        return None

    def list_local_files(self, resources_base_dir:str, ckan:CkanApi, cancel_if_present:bool=True) -> List[str]:
        # obtain row count of each resource
        self.local_file_size = [0] * len(self.resource_ids)
        for index, resource_id in enumerate(self.resource_ids):
            resource_info = ckan.get_resource_info_or_request(resource_id, error_not_found=False, datastore_info=True)
            if resource_info is not None and resource_info.datastore_info is not None and resource_info.datastore_info.row_count is not None:
                self.local_file_size[index] = resource_info.datastore_info.row_count
        self.local_file_size_sum = sum(self.local_file_size,0)
        return self.local_file_list

    def _update_metadata(self, ckan: CkanApi, *, base_dir:str=None):
        # obtain metadata from first resource of list
        super()._update_metadata(ckan, base_dir=base_dir)
        resource_info = ckan.get_resource_info_or_request(self.resource_ids[0], error_not_found=True, datastore_info=True)
        self.resource_attributes_data_source = None
        self.field_builders_data_source = None
        if resource_info is not None:
            self.resource_attributes_data_source = resource_info.copy()
            if resource_info.datastore_info is not None and resource_info.datastore_info.fields_dict is not None:
                self.field_builders_data_source = OrderedDict()
                for field_info in resource_info.datastore_info.fields_dict:
                    self.field_builders_data_source[field_info.field_name] = BuilderField._from_ckan_field(field_info)

    def get_local_df_chunk_generator(self, resources_base_dir:str, ckan:CkanApi, allow_chunks:bool=True,
                                     **kwargs) -> Generator[FileChunkDataFrame, None, None]:
        for file_index, resource_id in enumerate(self.resource_ids):
            self.file_semaphore.acquire()
            file_position = 0
            chunk_index = 0
            generator = ckan.datastore_search_page_generator(resource_id=resource_id, search_all=True)
            self.file_semaphore.release()
            while True:
                try:
                    self.file_semaphore.acquire()
                    df = next(generator)
                    chunk_index += 1
                    file_position += len(df)
                    self.read_line_counter += len(df)
                    line_counter = self.read_line_counter
                    self.file_semaphore.release()
                except StopIteration:
                    self.file_semaphore.release()
                    return
                yield FileChunkDataFrame(df, resource_id, file_index, chunk_index,
                                         file_position=file_position, read_line_counter=line_counter)





#!python3
# -*- coding: utf-8 -*-
"""
Harvester base class
"""
from typing import Union, List, Any, Callable, Tuple
from collections import OrderedDict
from abc import ABC, abstractmethod

import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage
from ckanapi_harvesters.harvesters.harvester_model import DatasetMetadata, TableMetadata
from ckanapi_harvesters.harvesters.harvester_params import DatabaseParams, DatasetParams, TableParams
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_abc import CkanDataCleanerABC


class HarvesterConnectABC(ABC):
    def __del__(self):
        self.disconnect()
        self.clear_secrets()

    @abstractmethod
    def clear_secrets(self):
        raise NotImplementedError()

    @abstractmethod
    def connect(self, *, cancel_if_connected:bool=True) -> Any:
        raise NotImplementedError()

    @abstractmethod
    def _finalize_connection(self):
        raise NotImplementedError()

    @abstractmethod
    def is_connected(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def check_connection(self, *, new_connection:bool=False, raise_error:bool=False) -> Union[None, ContextErrorLevelMessage]:
        raise NotImplementedError()

    @abstractmethod
    def disconnect(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def update_from_ckan(self, ckan):
        raise NotImplementedError()

    # Context Manager behavior ----------
    # to use Harvester in a "with" statement
    def __enter__(self):
        self.disconnect()  # start new sessions
        self.connect()  # connect at once
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.disconnect()
        self.clear_secrets()
        return True
    # -------------


class DatabaseHarvesterABC(HarvesterConnectABC, ABC):
    def __init__(self, params:DatabaseParams=None):
        if params is None:
            params = DatabaseParams()
        self.params: DatabaseParams = params

    def clear_secrets(self):
        if self.params.login is not None:
            self.params.login.clear()

    @abstractmethod
    def copy(self, *, dest=None):
        dest.params = self.params.copy()
        return dest

    @staticmethod
    @abstractmethod
    def init_from_options_string(options_string:str, *, base_dir:str=None) -> "DatabaseHarvesterABC":
        raise NotImplementedError()

    def _finalize_connection(self):
        pass

    ## query methods interface ---------------
    @abstractmethod
    def get_dataset_harvester(self, dataset_name:str) -> "DatasetHarvesterABC":
        raise NotImplementedError()

    @abstractmethod
    def list_datasets(self, return_metadata:bool=True) -> Union[List[str], OrderedDict[str, DatasetMetadata]]:
        raise NotImplementedError()

    def update_from_ckan(self, ckan):
        self.params._update_from_ckan(ckan)


class DatasetHarvesterABC(DatabaseHarvesterABC, ABC):
    def __init__(self, params:DatasetParams=None):
        if params is None:
            params = DatasetParams()
        super().__init__(params)
        self.params: DatasetParams = params
        self.dataset_metadata: Union[DatasetMetadata, None] = None

    @abstractmethod
    def _finalize_connection(self):
        raise NotImplementedError()

    @abstractmethod
    def copy(self, *, dest=None):
        super().copy(dest=dest)
        dest.dataset_metadata = self.dataset_metadata.copy() if self.dataset_metadata is not None else None
        return dest

    @staticmethod
    @abstractmethod
    def init_from_options_string(options_string:str, *, base_dir:str=None) -> "DatasetHarvesterABC":
        raise NotImplementedError()

    ## metadata interface ---------------
    @abstractmethod
    def query_dataset_metadata(self, cancel_if_present:bool=True) -> DatasetMetadata:
        self.connect()
        if cancel_if_present and self.dataset_metadata is not None:
            return self.dataset_metadata
        else:
            self.dataset_metadata = DatasetMetadata()
            # user needs to complete here
            self.dataset_metadata.tables = self.list_tables(return_metadata=True)
            return self.dataset_metadata

    def clean_dataset_metadata(self) -> DatasetMetadata:
        return self.query_dataset_metadata().copy()

    ## query methods interface ---------------
    @abstractmethod
    def get_table_harvester(self, table_name:str) -> "TableHarvesterABC":
        raise NotImplementedError()

    @abstractmethod
    def list_tables(self, return_metadata:bool=True) -> Union[List[str], OrderedDict[str, TableMetadata]]:
        raise NotImplementedError()


class TableHarvesterABC(DatasetHarvesterABC, ABC):
    _default_upload_fun: Union[Callable[[Any], pd.DataFrame], None] = None
    _default_primary_key: Union[List[str], None] = None

    def __init__(self, params:TableParams=None):
        if params is None:
            params = TableParams()
        super().__init__(params)
        self.params: TableParams = params
        self.table_metadata: Union[TableMetadata, None] = None

    @abstractmethod
    def copy(self, *, dest=None):
        super().copy(dest=dest)
        dest.table_metadata = self.table_metadata.copy() if self.table_metadata is not None else None
        return dest

    @staticmethod
    @abstractmethod
    def init_from_options_string(options_string:str, *, base_dir:str=None, file_url_attr:str=None) -> "TableHarvesterABC":
        raise NotImplementedError()

    ## metadata interface ---------------
    @abstractmethod
    def query_table_metadata(self, cancel_if_present:bool=True) -> TableMetadata:
        self.connect()
        if cancel_if_present and self.table_metadata is not None:
            return self.table_metadata
        else:
            self.table_metadata = TableMetadata()
            # user needs to complete here
            return self.table_metadata

    def clean_table_metadata(self) -> TableMetadata:
        return self.query_table_metadata().copy()

    @classmethod
    def get_default_df_upload_fun(cls) -> Union[Callable[[Any], pd.DataFrame], None]:
        return cls._default_upload_fun

    def get_default_data_cleaner(self) -> Union[CkanDataCleanerABC, None]:
        return None

    @abstractmethod
    def get_default_primary_key(self) -> List[str]:
        return []

    ## query methods interface ---------------
    @abstractmethod
    def list_queries(self, *, new_connection:bool=False) -> List[Tuple[Any,int]]:
        self.connect(cancel_if_connected=not new_connection)
        raise NotImplementedError()

    @abstractmethod
    def query_data(self, query:Any) -> Union[List[dict], pd.DataFrame]:
        raise NotImplementedError()


